/*****************************************************************************
 * $Id: greedyMap.hh,v 1.1 2006-02-25 05:39:54 devel Exp $
 * @file
 * @author	Jim E. Brooks  http://www.jimbrooks.org
 * @brief	C++ template for an associative container that finds the longest matching key.
 * @remarks This is a C++ template for an associative container with two
 *          special matching behaviors:
 *          - Greedy matching:
 *            Greedily searches for the longest key that matches a pattern.
 *          - Partial matching:
 *            Any key can match a pattern if the key is a subset of the leftmost
 *            chars in the pattern (but not v.v.).
 *          As an example of both greedy and partial matching,
 *          assuming { "ab", "abc", "abcd" } are stored in a GreedyMap,
 *          the given pattern "abc123" will be matched with the key "abc".
 *          Why?  First note that key "ab" matches the leftmost part of the pattern.
 *          But since we're greedy, matching continues to the key "abc" which
 *          also forms a partial match.  But the "d" in the key "abcd" causes
 *          a mismatch with the pattern.
 *//*
 * LEGAL:	COPYRIGHT (C) JIM E. BROOKS 2005
 *          THIS SOURCE CODE IS RELEASED UNDER THE TERMS
 *          OF THE GNU GENERAL PUBLIC LICENSE (GPL).
 *****************************************************************************/

#ifndef GREEDYMAP_HH
#define GREEDYMAP_HH 1

#include <string.h>

#include <string>
#include <vector>
#include <utility>
#include <algorithm>

////////////////////////////////////////////////////////////////////////////////
/// @brief GreedyMap
/// @verbatim
///
/// The GreedyMap is a C++ vector consisting of adjacent blocks of 256 pointers.
/// One block is allocated for every char in the longest key encountered so far.
/// The value of a key is referenced by a pointer in a block that correlates
/// to the last char of the key.  Like a tree, the value can be attained after
/// the end of a path is reached.
///
/// Possible values of VectorKeyVal pointers:
///
///	- NULL:
///	  soughtKey[i] doesn't match any char i in the stored keys.
///
/// - &emptyVectorKeyVal:
///	  soughtKey[i] matches but soughtKey[0:i+1] is not yet a complete match.
///	  No stored key has this char as its last char.
///
/// - A non-empty VectorKeyVal:
///	  Indicates a possible complete match and points to a vector of key,value pairs.
///	  The sought key is compared against every key in the vector (string comparison).
///
/// The GreedyMap is designed to solve these problems:
///
/// - The keys "cat" and "rat" are sought and were stored:
///
///   Both keys lead to the same destination block[2]['t']
///   but are disambiguated by a full string comparison.
///
///   block[0]['c'] = 1
///   block[0]['r'] = 1
///   block[1]['a'] = 1
///   block[2]['t'] = vector[ ("cat","feline"), ("rat","rodent") ]
///
/// - The keys "a0" and "b1" were stored then "a1" is sought:
///
///   Both keys lead to the same destination block[2]['t']
///   but are disambiguated by a full string comparison.
///
///   block[0]['a'] = 1
///   block[0]['b'] = 1
///   block[1]['0'] = vector[ ("a0",true) ]
///   block[1]['1'] = vector[ ("b1",true) ]
///
///   To find "a1", block[0]['a'], block[1]['1'] will be traversed.
///   But "a1" doesn't match "b1" in block[1]['1'].
///
/// @endverbatim
////////////////////////////////////////////////////////////////////////////////
template<class Val>
class GreedyMap
{
public:
            GreedyMap( unsigned int hintKeySize );
            ~GreedyMap();
    bool    Insert( const string& key, const Val& val, Val& valOld );
	bool	Find( const string& key, string& keyOut, Val& valOut );
	void	ForEach( void(*Callback)(const string& key, const Val& val) );

private:
	typedef pair<string,Val>	KeyVal;
	typedef vector<KeyVal>		VectorKeyVal;
	vector<VectorKeyVal*>		blocks;					// length is multiple of 256
	VectorKeyVal				emptyVectorKeyVal;
	unsigned int	LongestKey( void) { return blocks.size() >> 8; }
};

/*****************************************************************************
 * Construct a GreedyMap.
 * @param hintKeySize
 *        Estimated length of the longest possible key.
 *****************************************************************************/
template<class Val>
GreedyMap<Val>::GreedyMap( unsigned int hintKeySize )
{
	blocks.reserve( hintKeySize << 8 );					// pre-allocate vector for speed
	for ( unsigned int i = 0; i < blocks.size(); ++i )	// invalidate every element
		blocks[i] = NULL;
	return;
}

/*****************************************************************************
 * Destroy a GreedyMap.
 *****************************************************************************/
template<class Val>
GreedyMap<Val>::~GreedyMap()
{
	// Delete vectors of key,value pairs (but not the empty one!).
	// The pairs were copies.
	for ( unsigned int i = 0; i < blocks.size(); ++i )
		if ( blocks[i] != &emptyVectorKeyVal ) delete blocks[i];
	return;
}

/*****************************************************************************
 * Insert a value into the map.
 * @param  key
 * @param  val
 * @param  valOld
 * @return If old value was overwritten (displaced), returns it and true.
 *****************************************************************************/
template<class Val>
bool
GreedyMap<Val>::Insert( const string& key, const Val& val, Val& valOld )
{
	// Extend blocks vector if this key is longer than its predecessors.
	int delta = (key.size() << 8) - blocks.size();
	while ( delta > 0 ) { --delta; blocks.push_back( NULL ); }

	// For every char in the passed key except the last, store a pointer
	// to emptyVectorKeyVal into every correlating block element unless
	// its points to another VectorKeyVal.
	unsigned int i, idx;
	for ( i = 0; i < key.size()-1; ++i )	// for every char in key except the last
	{
		idx = (i << 8) + int(key[i]);
		if ( blocks[idx] == NULL ) blocks[idx] = &emptyVectorKeyVal;
	}

	// i indexes last char of key.
	// Store value in a vector of key,values.
	idx = (i << 8) + int(key[i]);
	if ( blocks[idx] == NULL || blocks[idx] == &emptyVectorKeyVal )
	{												// Hey Beavis, don't append to emptyVectorKeyVal
		blocks[idx] = new VectorKeyVal;
		blocks[idx]->push_back( KeyVal(key,val) );	// new pair with copies of key,val
		return false;
	}
	else
	{
		// If the same key exists, overwrite
		// the old value with the new.
		VectorKeyVal& vec = *blocks[idx];			// may point to emptyVectorKeyVal
		for ( unsigned int vi = 0; vi < vec.size(); ++vi )
		{
			if ( key == vec[vi].first )
			{
				// Overwrite.
				valOld         = vec[vi].second;
				vec[vi].first  = key;
				vec[vi].second = val;
				return true;
			}
		}
		// The new value is different.
		blocks[idx]->push_back( KeyVal(key,val) );	// new pair with copies of key,val
	}

	return false;
}

/*****************************************************************************
 * Greedily find the longest match in the map.
 * @return True if a match was found.
 * @param  key
 *         The key that is sought.
 * @param  keyOut
 *         The key that was matched else undefined.
 *         May or may not be the same as the input key.
 * @param  valOut
 *         Matched value else undefined.
 *****************************************************************************/
template<class Val>
bool
GreedyMap<Val>::Find( const string& key, string& keyOut, Val& valOut )
{
	// Stop whenever a NULL is encountered.
	// But greedily continue if a match is found to try to find the longest match.
	// If the sought key has chars at the right that don't match but all of
	// the preceding left chars do match, then that is treated as a match.

	keyOut = "";

	// For each char in sought key (but not more than the longest stored key).
	unsigned int cnt = min( (unsigned int)key.size(), LongestKey() );
	for ( unsigned int i = 0; i < cnt; ++i )
	{
		unsigned int idx = (i << 8) + int(key[i]);

		// If soughtKey[i] doesn't match any corresponding char in all stored keys.
		if ( blocks[idx] == NULL )
			break;

		// soughtKey[i] matches one or more corresponding chars in the stored keys,
		// but that doesn't mean the key is a match.  For example, if "a0" and "b1"
		// are the stored keys, then "a1" has chars that match chars in different
		// stored keys but it doesn't match a single stored key.
		VectorKeyVal& vec = *blocks[idx];		// may point to emptyVectorKeyVal (size=0)
		for ( unsigned int vi = 0; vi < vec.size(); ++vi )
		{
//			if ( key.compare( 0, i, vec[vi].first ) == 0 )
			if ( memcmp( key.c_str(), vec[vi].first.c_str(), i+1 ) == 0 )
			{
				// Found a match.
				keyOut = vec[vi].first;
				valOut = vec[vi].second;
				break;
			}
		}
	}

	return keyOut.size() > 0;
}

/*****************************************************************************
 * Invoke a callback for each stored key/value pair in the map.
 *****************************************************************************/
template<class Val>
void
GreedyMap<Val>::ForEach( void(*Callback)(const string& key, const Val& val) )
{
	for ( unsigned int idx = 0; idx < blocks.size(); ++idx )
	{
		// If this points to a VectorKeyVal.
		if ( blocks[idx] != NULL )
		{
			VectorKeyVal& vec = *blocks[idx];
			for ( unsigned int vi = 0; vi < vec.size(); ++vi )
				Callback( vec[vi].first, vec[vi].second );
		}
	}
	return;
}

#endif // GREEDYMAP_HH
