/*****************************************************************************
 * $Id: main.cc,v 1.9 2006-02-25 05:39:54 devel Exp $
 * @file
 * @author	Jim E. Brooks  http://www.jimbrooks.org
 * @brief	C++ component of pypersrc.
 * @remarks pypersrc is written in Python and C++.
 *          pypersrc is a C++ executable that both embeds and extends Python.
 *//*
 * LEGAL:	COPYRIGHT (C) JIM E. BROOKS 2005
 *          THIS SOURCE CODE IS RELEASED UNDER THE TERMS
 *          OF THE GNU GENERAL PUBLIC LICENSE (GPL).
 *****************************************************************************/

#include <Python.h>
#include "common.hh"
#include "ctags_ifc.h"

// Python state.
Python python;

/*****************************************************************************
 * main.
 *****************************************************************************/
int
main( int argc, char** argv )
{
	string func         = "Startup";		// Python names
	string importModule = "pypersrc";
	string exportModule = "hypersrc_cpp";
    bool   testing      = false;

	// Parse cmd-line args.
    // Build a vector of args intended for Python minus args handled here.
	vector<string> args;
	for ( int i = 0; i < argc; ++i )
	{
		if ( i == 0 )
        {
			args.push_back( importModule ); // let argv[0] be Python module name

			// Hidden arg to tell Python code its version.
			args.push_back( string("-setversion") );
			args.push_back( string(PYPERSRC_VERSION_STR) );
        }
		else if ( STREQ( argv[i], "-test" ) )
		{
            testing = true;

            // Prepare to import test Python module from tests/ dir.
			importModule = args[0] = argv[i+1];
			cerr << "[C++]: Testing Python module " << importModule << endl;

            // Append PYTHON_PATH the path to tests/ dir.
        	string execPath;
        	if ( GetExecPath( argv[0], execPath ) )
            {
                string testPath = execPath + "/../tests";
                AppendPythonPath( testPath );
            }

			++i;
		}
		else if ( STREQ( argv[i], "-version" ) || STREQ( argv[i], "--version" ) ) // not -v which means verbose
		{
			cout << PYPERSRC_VERSION_STR << endl;
			Exit(0);
		}
        else
        {
			args.push_back( string(argv[i]) );
        }
	}

    // Initialize Exuberant ctags.
    // -x : Print a tabular, human-readable cross reference (xref) file
    // -u : don't sort tags
    static const char* argsCtags[] = { "ctags", "-x", "-u" };
    CtagsStart( sizeof(argsCtags)/sizeof(char*), (char**)argsCtags );

	// Initialize Python.
	if ( ! python.Init( argc, const_cast<const char**>(argv), importModule ) )
	{
		cerr << "ERROR: Can't initialize Python. \n";
        if ( testing )
            cerr << "Try passing only the basename of a test in the tests/ directory. \n";
		Exit(1);
	}

	// Call Python function.
	if ( ! python.Call( func, args ) )
	{
		cerr << "ERROR: can't call Python function " << func << "()\n";
		Exit(1);
	}

	Exit(0);
	return 0;
}

/*****************************************************************************
 * Initialize embedded Python.
 * This imports a Python module from Python to C++ and v.v exports a Python module
 * from C++ to Python (embeds and extends Python).
 * @param	argv
 *          From cmd-line.
 * @param	moduleImport
 *          Name of module to import from Python to C++.
 * @param	moduleExport
 *          Name of module to export to Python from C++.
 * @param	methodsExport
 *          Table of C++ functions to export to Python as Python methods.
 *****************************************************************************/
bool
Python::Init( int argc, const char** argv, const string& importModule )
{
	mInit = false;

	// Set or append PYTHONPATH.
	string execPath;
	if ( GetExecPath( argv[0], execPath ) )
        AppendPythonPath( execPath );

	// Initialize Python.
	Py_Initialize();
	if ( ! Py_IsInitialized() )
    {
        cerr << "ERROR: Py_IsInitialized() failed. \n";
		return false;
    }
    PySys_SetArgv( argc, const_cast<char**>(argv) );

	// Export C++ functions as Python modules.
	if ( ! ExportAllModules() )
	{
        cerr << "ERROR: export C++ function to Python. \n";
		return false;
	}

	// Import module from Python to C++.
	mPyImportModule = PyImport_ImportModule( const_cast<char*>(importModule.c_str()) );
	if ( mPyImportModule == NULL )
    {
        cerr << "ERROR: PyImport_ImportModule('" << importModule << "') failed. \n";
		return false;
    }

	// Prepare Python::Call().
	mPyImportModuleDict = PyModule_GetDict( mPyImportModule );
	if ( mPyImportModuleDict == NULL )
    {
        cerr << "ERROR: PyModule_GetDict() failed. \n";
		return false;
    }

	return (mInit = true);
}

/*****************************************************************************
 * Export all Python modules.
 *****************************************************************************/
bool
Python::ExportAllModules( void )
{
	extern PyMethodDef exportMethodsTags[];
	extern PyMethodDef exportMethodsDequeChar[];
	extern PyMethodDef exportMethodsGreedyDict[];
	extern PyMethodDef exportMethodsUtils[];
	return ExportModule( "pycpp_tags",       exportMethodsTags )
	    && ExportModule( "pycpp_dequechar",  exportMethodsDequeChar )
	    && ExportModule( "pycpp_greedydict", exportMethodsGreedyDict )
	    && ExportModule( "pycpp_utils",      exportMethodsUtils );
}

/*****************************************************************************
 * Export a Python module from C++ to Python (embeds and extends Python).
 * @param	moduleExport
 *          Name of module to export to Python from C++.
 * @param	methodsExport
 *          Table of C++ functions to export to Python as Python methods.
 *****************************************************************************/
bool
Python::ExportModule( const char* exportModule, PyMethodDef exportMethods[] )
{
	// Export C++ functions to Python.
	// Returned PyObject is a borrowed reference (don't DEC it).
	if ( Py_InitModule( exportModule, exportMethods ) == NULL )
		return false;
	else
		return true;
}

/*****************************************************************************
 * Call a Python function.
 * @param	func
 *          Name of Python function (in previously imported module) to call.
 * @param	args
 *****************************************************************************/
bool
Python::Call( const string& func, const vector<string>& args )
{
	bool res = false;
	PyObject* pyFunc		= NULL;
	PyObject* pyArgTuple	= NULL;
	PyObject* pyResult		= NULL;
	vector<string>::const_iterator itr;
	int i;

	// Prepare to call a Python.
	// First step is to find it in the module dictionary.
	// The dictionary contains names of objects in the module.
	pyFunc = PyDict_GetItemString( mPyImportModuleDict, const_cast<char*>(func.c_str()) );
	if ( pyFunc == NULL || ! PyCallable_Check(pyFunc) )
		goto Error;

	// Next construct a Python tuple to hold the args to pass.
	// The args are from the cmd-line.
	pyArgTuple = PyTuple_New( args.size() );
	if ( pyArgTuple == NULL )
		goto Error;
	i = 0;
	for ( itr = args.begin(); itr != args.end(); ++itr )
	{
		PyTuple_SetItem( pyArgTuple, i++, PyString_FromString( (*itr).c_str() ) );
	}

	// Now call the Python function.
	mCalling = true;
	pyResult = PyObject_CallObject( pyFunc, pyArgTuple );
	mCalling = false;
	if ( pyResult == NULL )
		goto Error;
	res = true;
Return:
  //if ( pyFunc )     { Py_DECREF( pyFunc ); }     // borrowed reference
	if ( pyArgTuple ) { Py_DECREF( pyArgTuple ); }
	if ( pyResult )   { Py_DECREF( pyResult ); }
	return res;
Error:
	res = false;
	goto Return;
}

/*****************************************************************************
 * Stop Python.
 *****************************************************************************/
Python::~Python()
{
#define PY_DEC_NULL( OBJ ) if ( OBJ != NULL ) { Py_DECREF( OBJ ); OBJ = NULL; }
#define PY_NULL( OBJ )     OBJ = NULL;  // for borrowed references

	if ( mInit )
	{
		PY_DEC_NULL( mPyImportModule )
		PY_NULL(     mPyImportModuleDict ) // borrowed reference
		Py_Finalize();
		mInit = false;
	}
	return;
}

/*****************************************************************************
 * Append PYTHON_PATH environment variable.
 *****************************************************************************/
void
AppendPythonPath( const string& path )
{
	string pythonpath;
	const char* PYTHONPATH = getenv( "PYTHONPATH" );
	if ( PYTHONPATH != NULL )
    {
//		pythonpath = PYTHONPATH + ':' + path; // append     // WRONG FYI
		pythonpath = PYTHONPATH;
        pythonpath += ':';
        pythonpath += path; // append
    }
	else
    {
		pythonpath = path; // set
    }
    if ( pythonpath[pythonpath.size()-1] == '/' )
    	pythonpath.resize( pythonpath.size()-1 ); // chop off trailing slash
	if (setenv( "PYTHONPATH", pythonpath.c_str(), 1 ) == -1)
        cerr << "ERROR: setenv() failed \n";
    return;
}

/*****************************************************************************
 * Exit.
 *****************************************************************************/
void
Exit( int exitCode )
{
    // Exuberant ctags cleanup.
    CtagsEnd();

	// ~Python() will be called implicitly by the static python.
	exit( exitCode );
}
