/*****************************************************************************
 * $Id: pyext_DequeChar.cc,v 1.1 2006-02-25 05:39:54 devel Exp $
 * @file
 * @author	Jim E. Brooks  http://www.jimbrooks.org
 * @brief	DequeChar Python extension built over C++ deque<char>.
 * @remarks Python extension functions have the form:
 *          PyExt(PyObject* self, PyObject* args)
 *          I write PyExt(nil,args) because in fact self will be NULL
 *          (except for the esoteric "Python extension types").
 *          A few of my extensions need self which my Python code passes as
 *          the first arg which appears on the C++ side as the first element
 *          in the tuple passed as "args" to PyExt(self,args).
 *//*
 * LEGAL:	COPYRIGHT (C) JIM E. BROOKS 2005
 *          THIS SOURCE CODE IS RELEASED UNDER THE TERMS
 *          OF THE GNU GENERAL PUBLIC LICENSE (GPL).
 *****************************************************************************/

#include <Python.h>
#include "common.hh"

// Table of C++ function that act as Python methods (Python extensions).
PyMethodDef exportMethodsDequeChar[] =
{
    { "DequeChar_New",		DequeChar_New,		METH_VARARGS, "Create a new DequeChar C++ object." },
    { "DequeChar_Delete",	DequeChar_Delete,	METH_VARARGS, "Delete a DequeChar C++ object." },
    { "DequeChar_PushHead",	DequeChar_PushHead,	METH_VARARGS, "Prepend a Python string to a C++ DequeChar object." },
    { "DequeChar_PushTail",	DequeChar_PushTail,	METH_VARARGS, "Append a Python string to a C++ DequeChar object." },
    { "DequeChar_PopHead",	DequeChar_PopHead,	METH_VARARGS, "Pop/remove a Python string from head of C++ DequeChar object." },
    { "DequeChar_PopTail",	DequeChar_PopTail,	METH_VARARGS, "Pop/remove a Python string from tail of C++ DequeChar object." },
    { "DequeChar_PeekHead",	DequeChar_PeekHead,	METH_VARARGS, "Peek a Python string at head of C++ DequeChar object." },
    { "DequeChar_PeekTail",	DequeChar_PeekTail,	METH_VARARGS, "Peek a Python string at tail of C++ DequeChar object." },
    { "DequeChar_IfEmpty",	DequeChar_IfEmpty,	METH_VARARGS, "Return True if C++ DequeChar object is empty." },
    { "DequeChar_Size",     DequeChar_Size,     METH_VARARGS, "Return amount of elements in C++ DequeChar object." },
    { "DequeChar_Set",      DequeChar_Set,      METH_VARARGS, "Set contents of C++ DequeChar object from a Python string." },
    { "DequeChar_Get",      DequeChar_Get,      METH_VARARGS, "Return contents of C++ DequeChar object as a Python string." },
    { "DequeChar_SetItem",  DequeChar_SetItem,  METH_VARARGS, "Assign to a char by index in C++ DequeChar object." },
    { "DequeChar_GetItem",  DequeChar_GetItem,  METH_VARARGS, "Return char by index from C++ DequeChar object." },
    { "DequeChar_Index",    DequeChar_Index,    METH_VARARGS, "Return index of substring in C++ DequeChar object." },
	{ NULL, NULL, 0, NULL }
};

// Associates a DequeChar C++ object with a DequeChar Python object.
static DequeCharMap sDequeCharMap;

// If Python object is associated with a C++ DequeChar object.
#define IF_PY_OBJECT_MAPPED(ITR)        								\
	DequeCharMap::const_iterator ITR = sDequeCharMap.find( self );		\
	if ( ITR != sDequeCharMap.end() )

/*****************************************************************************
 * Interface to Python: Create a new DequeChar C++ object.
 *****************************************************************************/
PyObject*
DequeChar_New( PyObject* nil, PyObject* args )
{
	PyObject*	self;
    if ( PyArg_ParseTuple( args, "O", &self ) )
	{
		// Create a DequeChar C++ object and associate with a Python self object.
		DequeChar* dequeChar = new DequeChar();
		sDequeCharMap.insert( make_pair<PyObject*,DequeChar*>(self,dequeChar) );
	}
	PY_RETURN_NONE
}

/*****************************************************************************
 * Interface to Python: Delete a DequeChar C++ object.
 *****************************************************************************/
PyObject*
DequeChar_Delete( PyObject* nil, PyObject* args )
{
	PyObject*	self;
    if ( PyArg_ParseTuple( args, "O", &self ) )
	{
		IF_PY_OBJECT_MAPPED(itr)
		{
			// Delete DequeChar C++ object.
			DequeChar* dequeChar = itr->second;
			delete dequeChar;
			sDequeCharMap.erase( self );
		}
	}
	PY_RETURN_NONE
}

/*****************************************************************************
 * Interface to Python: Prepend a Python string to a C++ DequeChar object.
 *****************************************************************************/
PyObject*
DequeChar_PushHead( PyObject* nil, PyObject* args )
{
	PyObject*   self;
	const char* chars;
	int         charsLen;
    if ( PyArg_ParseTuple( args, "Os#", &self, &chars, &charsLen ) )
	{
		IF_PY_OBJECT_MAPPED(itr)
		{
			DequeChar* dequeChar = itr->second;
			for ( int i = charsLen-1; i >= 0; --i )
			{
				dequeChar->push_front( chars[i] );
			}
			PY_RETURN_NONE
		}
	}
	PY_RETURN_BAD_OBJECT_ERROR
}

/*****************************************************************************
 * Interface to Python: Append a Python string to a C++ DequeChar object.
 *****************************************************************************/
PyObject*
DequeChar_PushTail( PyObject* nil, PyObject* args )
{
	PyObject*   self;
	const char* chars;
	int         charsLen;
    if ( PyArg_ParseTuple( args, "Os#", &self, &chars, &charsLen ) )
	{
		IF_PY_OBJECT_MAPPED(itr)
		{
			DequeChar* dequeChar = itr->second;
			for ( int i = 0; i < charsLen; ++i )
				dequeChar->push_back( chars[i] );
			PY_RETURN_NONE
		}
	}
	PY_RETURN_BAD_OBJECT_ERROR
}

/*****************************************************************************
 * Interface to Python: Pop/remove a Python string from head of C++ DequeChar object.
 * Raises exception if not all n chars are available.
 *****************************************************************************/
PyObject*
DequeChar_PopHead( PyObject* nil, PyObject* args )
{
	PyObject*	self;
	int			n;
    if ( PyArg_ParseTuple( args, "Oi", &self, &n ) )
	{
		IF_PY_OBJECT_MAPPED(itr)
		{
  			DequeChar* dequeChar = itr->second;
            if ( n > 0  &&  n <= (int)dequeChar->size() )
            {
        		string chars;
                for ( ; n; n-- )
                {
                    char c = dequeChar->front(); dequeChar->pop_front();
                    chars += c;
                }
                return Py_BuildValue( "s#", chars.c_str(), chars.size() );
            }
			PY_RETURN_INDEX_ERROR
		}
	}
	PY_RETURN_BAD_OBJECT_ERROR
}

/*****************************************************************************
 * Interface to Python: Pop/remove a Python string from tail of C++ DequeChar object.
 * Raises exception if not all n chars are available.
 *****************************************************************************/
PyObject*
DequeChar_PopTail( PyObject* nil, PyObject* args )
{
	PyObject*	self;
	int			n;
    if ( PyArg_ParseTuple( args, "Oi", &self, &n ) )
	{
		IF_PY_OBJECT_MAPPED(itr)
		{
            DequeChar* dequeChar = itr->second;
            if ( n > 0  &&  n <= (int)dequeChar->size() )
            {
                string chars;
                for ( ; n; n-- )
                {
                    char c = dequeChar->back();	dequeChar->pop_back();
                    //chars.insert( 0, c );
                    chars.insert( (string::size_type)0, 1, c );  // idx, n, c
                }
                return Py_BuildValue( "s#", chars.c_str(), chars.size() );
            }
			PY_RETURN_INDEX_ERROR
		}
	}
	PY_RETURN_BAD_OBJECT_ERROR
}

/*****************************************************************************
 * Interface to Python: Peek a Python string at head of C++ DequeChar object.
 * Raises exception if not all n chars are available.
 *****************************************************************************/
PyObject*
DequeChar_PeekHead( PyObject* nil, PyObject* args )
{
	PyObject*	self;
	int			n;
    if ( PyArg_ParseTuple( args, "Oi", &self, &n ) )
	{
		IF_PY_OBJECT_MAPPED(itr)
		{
            DequeChar* dequeChar = itr->second;
            if ( n > 0  &&  n <= (int)dequeChar->size() )
            {
                string chars;
    			for ( int i = 0; i < n; ++i )
    				chars += (*dequeChar)[i];
    			return Py_BuildValue( "s#", chars.c_str(), n );
            }
            PY_RETURN_INDEX_ERROR
		}
	}
	PY_RETURN_BAD_OBJECT_ERROR
}

/*****************************************************************************
 * Interface to Python: Peek a Python string at tail of C++ DequeChar object.
 * Raises exception if not all n chars are available.
 *****************************************************************************/
PyObject*
DequeChar_PeekTail( PyObject* nil, PyObject* args )
{
	PyObject*	self;
	int			n;
    if ( PyArg_ParseTuple( args, "Oi", &self, &n ) )
	{
		IF_PY_OBJECT_MAPPED(itr)
		{
            DequeChar* dequeChar = itr->second;
            if ( n > 0  &&  n <= (int)dequeChar->size() )
            {
                string chars;
    			for ( uint i = dequeChar->size() - n; i < dequeChar->size(); ++i )
        			chars += (*dequeChar)[i];
    			return Py_BuildValue( "s#", chars.c_str(), n );
            }
            PY_RETURN_INDEX_ERROR
		}
	}
	PY_RETURN_BAD_OBJECT_ERROR
}

/*****************************************************************************
 * Interface to Python: Return True if C++ DequeChar object is empty.
 *****************************************************************************/
PyObject*
DequeChar_IfEmpty( PyObject* nil, PyObject* args )
{
	PyObject*	self;
    if ( PyArg_ParseTuple( args, "O", &self ) )
	{
		IF_PY_OBJECT_MAPPED(itr)
		{
			DequeChar* dequeChar = itr->second;
			return Py_BuildValue( "i", dequeChar->empty() );
		}
	}
	PY_RETURN_BAD_OBJECT_ERROR
}

/*****************************************************************************
 * Interface to Python: Return amount of elements in C++ DequeChar object.
 *****************************************************************************/
PyObject*
DequeChar_Size( PyObject* nil, PyObject* args )
{
	PyObject*	self;
    if ( PyArg_ParseTuple( args, "O", &self ) )
	{
		IF_PY_OBJECT_MAPPED(itr)
		{
			DequeChar* dequeChar = itr->second;
			return Py_BuildValue( "i", dequeChar->size() );
		}
	}
	PY_RETURN_BAD_OBJECT_ERROR
}

/*****************************************************************************
 * Interface to Python: Set contents of C++ DequeChar object from a Python string.
 * Also can be used to clear the deque by passing an empty string.
 *****************************************************************************/
PyObject*
DequeChar_Set( PyObject* nil, PyObject* args )
{
	PyObject*   self;
	const char* chars;
	int         charsLen;
    if ( PyArg_ParseTuple( args, "Os#", &self, &chars, &charsLen ) )
	{
		IF_PY_OBJECT_MAPPED(itr)
		{
			DequeChar* dequeChar = itr->second;
            dequeChar->clear();
            for ( int i = 0; i < charsLen; ++i )
                dequeChar->push_back( chars[i] );
			PY_RETURN_NONE
		}
	}
	PY_RETURN_BAD_OBJECT_ERROR
}

/*****************************************************************************
 * Interface to Python: Return contents of C++ DequeChar object as a Python string.
 *****************************************************************************/
PyObject*
DequeChar_Get( PyObject* nil, PyObject* args )
{
	PyObject*	self;
    if ( PyArg_ParseTuple( args, "O", &self ) )
	{
		IF_PY_OBJECT_MAPPED(itr)
		{
			DequeChar* dequeChar = itr->second;
			if ( ! dequeChar->empty() )
			{
				return PythonStringFromDequeChar( dequeChar, dequeChar->begin(), dequeChar->size() );
			}
			else
			{
				PY_RETURN_INDEX_ERROR
			}
		}
	}
	PY_RETURN_BAD_OBJECT_ERROR
}

/*****************************************************************************
 * Interface to Python: Assign to a char by index in C++ DequeChar object.
 * A negative index within range is allowed.
 * Raises exception if index is outside range (this isn't meant to extend the deque).
 *****************************************************************************/
PyObject*
DequeChar_SetItem( PyObject* nil, PyObject* args )
{
	PyObject*	self;
    int         idx;
    char        c;
    if ( PyArg_ParseTuple( args, "Oic", &self, &idx, &c ) )
	{
		IF_PY_OBJECT_MAPPED(itr)
		{
			DequeChar* dequeChar = itr->second;
			int dequeCharSize = dequeChar->size();
			// deque[-size] is allowed (equivalent to deque[0])
			// but deque[size] is one too much.
	        if (    (idx >= 0 && idx <   dequeCharSize)
                 || (idx <  0 && idx >= -dequeCharSize) )
			{
				if ( idx < 0 ) idx = dequeCharSize + idx;   // convert negative index
				(*dequeChar)[idx] = c;
                PY_RETURN_NONE
			}
			else
			{
				PY_RETURN_INDEX_ERROR
			}
		}
	}
	PY_RETURN_BAD_OBJECT_ERROR
}

/*****************************************************************************
 * Interface to Python: Return a char by index from C++ DequeChar object.
 * A negative index within range is allowed.
 *****************************************************************************/
PyObject*
DequeChar_GetItem( PyObject* nil, PyObject* args )
{
	PyObject*	self;
    int         idx;
    if ( PyArg_ParseTuple( args, "Oi", &self, &idx ) )
	{
		IF_PY_OBJECT_MAPPED(itr)
		{
			DequeChar* dequeChar = itr->second;
			int dequeCharSize = dequeChar->size();
			// deque[-size] is allowed (equivalent to deque[0])
			// but deque[size] is one too much.
	        if (    (idx >= 0 && idx <   dequeCharSize)
                 || (idx <  0 && idx >= -dequeCharSize) )
			{
				if ( idx < 0 ) idx = dequeCharSize + idx;   // convert negative index
				return Py_BuildValue( "c", (*dequeChar)[idx] );
			}
			else
			{
				PY_RETURN_INDEX_ERROR
			}
		}
	}
	PY_RETURN_BAD_OBJECT_ERROR
}

/*****************************************************************************
 * Interface to Python: Return index of substring in C++ DequeChar object.
 * Raises exception if no match, sought string is empty,
 * or index plus sought string is past end of deque.
 *****************************************************************************/
PyObject*
DequeChar_Index( PyObject* nil, PyObject* args )
{
	PyObject*	self;
	const char* chars;
	int			charsLen;
	int			idxStart;	// where to begin search
    if ( PyArg_ParseTuple( args, "Os#i", &self, &chars, &charsLen, &idxStart ) )
	{
		IF_PY_OBJECT_MAPPED(itr)
		{
			DequeChar* dequeChar = itr->second;
			if (    idxStart >= 0
                 && idxStart + charsLen <= (int)dequeChar->size()
                 && charsLen > 0 )
			{
				DequeChar::const_iterator itrDeq;
				string sought( chars, charsLen );
				itrDeq = FindSubset<DequeChar,string>( dequeChar->begin()+idxStart, dequeChar->end(),
	                                                   sought.begin(), sought.end() );
				if ( itrDeq != dequeChar->end() )
				{
                    int idx = itrDeq - dequeChar->begin();
                    return Py_BuildValue( "i", idx );
				}
			}
			PY_RETURN_INDEX_ERROR
		}
	}
	PY_RETURN_BAD_OBJECT_ERROR
}

/*****************************************************************************
 * Return a Python string from a range in a DequeChar.
 *****************************************************************************/
PyObject*
PythonStringFromDequeChar( const DequeChar* dequeChar, DequeChar::const_iterator itr, int len )
{
	// This is a downside to using a deque.  Casting a range in the deque as char*
    // would be ideal but that isn't possible since a deque is discontiguous.
    // So the string can only be made char-by-char.
	string chars;
	for ( int i = 0; i < len; ++i,++itr ) chars += *itr;
	return Py_BuildValue( "s#", chars.c_str(), chars.size() );
}
