/*****************************************************************************
 * $Id: pyext_GreedyDict.cc,v 1.1 2006-02-25 05:39:54 devel Exp $
 * @file
 * @author	Jim E. Brooks  http://www.jimbrooks.org
 * @brief	GreedyDict Python extension built over C++ GreedyMap template class.
 * @remarks GreedyDict and GreedyMap are the same concepts.
 * ////////////////////////////////////////////////////////////////////////////////
 * // GreedyDict and GreedyMap are the same concepts, but the names differ
 * // to follow Python and C++ naming conventions, resp.
 * // GreedyMap is a C++ template class (GreedyMap.hh).
 * // GreedyDict is a typedef for GreedyMap<PyObject*> and was given this name
 * // as it closely correlates to the GreedyDict Python class.
 * ////////////////////////////////////////////////////////////////////////////////
 *          Python extension functions have the form:
 *          PyExt(PyObject* self, PyObject* args)
 *          I write PyExt(nil,args) because in fact self will be NULL
 *          (except for the esoteric "Python extension types").
 *          A few of my extensions need self which my Python code passes as
 *          the first arg which appears on the C++ side as the first element
 *          in the tuple passed as "args" to PyExt(self,args).
 *//*
 * LEGAL:	COPYRIGHT (C) JIM E. BROOKS 2005
 *          THIS SOURCE CODE IS RELEASED UNDER THE TERMS
 *          OF THE GNU GENERAL PUBLIC LICENSE (GPL).
 *****************************************************************************/

#include <Python.h>
#include "common.hh"

// Table of C++ function that act as Python methods (Python extensions).
PyMethodDef exportMethodsGreedyDict[] =
{
    { "GreedyDict_New",		GreedyDict_New,		METH_VARARGS, "Create a new GreedyDict C++ object." },
    { "GreedyDict_Delete",	GreedyDict_Delete,	METH_VARARGS, "Delete a GreedyDict C++ object." },
    { "GreedyDict_Insert",	GreedyDict_Insert,	METH_VARARGS, "Insert a key/value pair into a GreedyDict." },
    { "GreedyDict_Find",	GreedyDict_Find,	METH_VARARGS, "Greedily find the longest match." },
	{ NULL, NULL, 0, NULL }
};

// Associates a GreedyMap C++ object () with a GreedyDict Python object.
static GreedyDictMap sGreedyDictMap;

// If Python object is associated with a C++ GreedyMap object.
#define IF_PY_OBJECT_MAPPED(ITR)        								\
	GreedyDictMap::const_iterator ITR = sGreedyDictMap.find( self );	\
	if ( ITR != sGreedyDictMap.end() )

/*****************************************************************************
 * Interface to Python: Create a new GreedyDict C++ object.
 *****************************************************************************/
PyObject*
GreedyDict_New( PyObject* nil, PyObject* args )
{
	PyObject*	self;
	int			hintKeySize;
    if ( PyArg_ParseTuple( args, "Oi", &self, &hintKeySize ) )
	{
		// Create a GreedyDict C++ object and associate with a Python self object.
		GreedyDict* greedyDict = new GreedyDict( hintKeySize );
		sGreedyDictMap.insert( make_pair<PyObject*,GreedyDict*>(self,greedyDict) );
	}
	PY_RETURN_NONE
}

/*****************************************************************************
 * Interface to Python: Delete a GreedyDict C++ object.
 *****************************************************************************/
// Second arg is tricky to declare: it's a reference to a const-pointer, not pointer-to-const.
void
GreedyDictDeleteCallback( const string& key, PyObject* const& val )
{
	// Decrement reference to this Python object.
	Py_DECREF(val);
}
PyObject*
GreedyDict_Delete( PyObject* nil, PyObject* args )
{
	PyObject*	self;
    if ( PyArg_ParseTuple( args, "O", &self ) )
	{
		IF_PY_OBJECT_MAPPED(itr)
		{
			GreedyDict* greedyDict = itr->second;

			// Decrement the reference counts of all Python objects
			// that were inserted in the GreedyDict so that Python's
			// garbage collector can free them.
			greedyDict->ForEach( GreedyDictDeleteCallback );

			// Now delete the GreedyDict and disassociate the Python object.
			delete greedyDict;
			sGreedyDictMap.erase( self );
		}
	}
	PY_RETURN_NONE
}

/*****************************************************************************
 * Interface to Python: Insert a key/value pair into a GreedyDict.
 *****************************************************************************/
PyObject*
GreedyDict_Insert( PyObject* nil, PyObject* args )
{
	PyObject*   self;
	const char* key;
	PyObject*   val;
    if ( PyArg_ParseTuple( args, "OsO", &self, &key, &val ) )
	{
		IF_PY_OBJECT_MAPPED(itr)
		{
			GreedyDict* greedyDict = itr->second;
			PyObject*   valOld;
			if ( greedyDict->Insert( string(key), val, valOld ) )
            {
				Py_DECREF(valOld);	// old Python object was displaced
            }
			Py_INCREF(val);			// this Python object is referenced while inside the map
			PY_RETURN_NONE
		}
	}
	PY_RETURN_BAD_OBJECT_ERROR
}

/*****************************************************************************
 * Interface to Python: Greedily find the longest match.
 * Returns a tuple (key,value) if found or raises an IndexError exception.
 *****************************************************************************/
PyObject*
GreedyDict_Find( PyObject* nil, PyObject* args )
{
	PyObject*   self;
	const char* key;
    if ( PyArg_ParseTuple( args, "Os", &self, &key ) )
	{
		IF_PY_OBJECT_MAPPED(itr)
		{
			GreedyDict* greedyDict = itr->second;
			string     keyMatch;
			PyObject*  valMatch;
			if ( greedyDict->Find( key, keyMatch, valMatch ) )
			{
				return Py_BuildValue( "(sO)", keyMatch.c_str(), valMatch );
			}
			else
			{
				PY_RETURN_INDEX_ERROR
			}
		}
	}
	PY_RETURN_BAD_OBJECT_ERROR
}
