/*****************************************************************************
 * $Id: pyext_Tags.cc,v 1.1 2006-02-25 05:39:54 devel Exp $
 * @file
 * @author	Jim E. Brooks  http://www.jimbrooks.org
 * @brief	C++ component of pypersrc.
 * @remarks Tags are stored in a C++ multimap which supports non-unique keys.
 *          A Python dictionary can't be used since its keys must be unique.
 *          To support case-insensitively finding a tag, a tag's all-lowercase
 *          name is used as the key and its original name is stored in the value.
 *
 *          Python extension functions have the form:
 *          PyExt(PyObject* self, PyObject* args)
 *          I write PyExt(nil,args) because in fact self will be NULL
 *          (except for the esoteric "Python extension types").
 *          A few of my extensions need self which my Python code passes as
 *          the first arg which appears on the C++ side as the first element
 *          in the tuple passed as "args" to PyExt(self,args).
 *//*
 * LEGAL:	COPYRIGHT (C) JIM E. BROOKS 2005
 *          THIS SOURCE CODE IS RELEASED UNDER THE TERMS
 *          OF THE GNU GENERAL PUBLIC LICENSE (GPL).
 *****************************************************************************/

#include <Python.h>
#include "common.hh"
#include "ctags_ifc.h"  // Exuberant ctags interface

#include <sys/types.h>
#include <regex.h>

// Table of C++ function that act as Python methods (Python extensions).
PyMethodDef exportMethodsTags[] =
{
    { "TagFile",			TagFile,			METH_VARARGS, "Tag symbols in a src file." },
    { "TagCnt",				TagCnt,				METH_VARARGS, "Return amount of tags." },
    { "InsertTag",			InsertTag,			METH_VARARGS, "Insert tags into C++ multimap." },
    { "FreeTags",			FreeTags,			METH_VARARGS, "Removes tags belonging to file from C++ multimap." },
    { "IfTag",				IfTag,				METH_VARARGS, "Return true if this is a tag." },
    { "IfTagRef",			IfTagRef,			METH_VARARGS, "Return true if this is reference to a tag." },
    { "FindTag",			FindTag,			METH_VARARGS, "Return list of matching tags or None." },
    { "GetModuleTags",		GetModuleTags,		METH_VARARGS, "Return list of tags in a module or None." },
	{ NULL, NULL, 0, NULL }
};

////////////////////////////////////////////////////////////////////////////////
/// @brief Tags class.
class Tags
{
public:
	typedef short TagInt; // types and files are enumerated to save memory
	class TagVal
	{
	public:
		TagVal( const string& name, TagInt type, TagInt file, TagInt line )
		{
            mName = name;
			mType = type;
			mFile = file;
			mLine = line;
		}
        string  mName;
		TagInt	mType;
		TagInt	mFile;
		TagInt	mLine;
	};

	// TagMap associates a tag name to a source code tag made by Exuberant ctags.
	// The key is stored as lower-case for case-insensitive searching.
	// Case-sensitive searching can be done by matching TagVal.mName
	// which holds the original case.
	typedef multimap<string,TagVal> TagMap;
	TagMap	mTagMap;

	// These help conserve memory by allowing a short int
	// to be stored in TagMap in lieue of a string.
	typedef map<string,TagInt>	TypeStrNumMap;
	typedef map<TagInt,string>	TypeNumStrMap;	// reverse map

	typedef map<string,TagInt>	FileStrNumMap;
	typedef map<TagInt,string>	FileNumStrMap;	// reverse map

	TypeStrNumMap	mTypeStrNumMap;
	TypeNumStrMap	mTypeNumStrMap;
	TagInt			mTypeNum;

	FileStrNumMap	mFileStrNumMap;
	FileNumStrMap	mFileNumStrMap;
	TagInt			mFileNum;

	// Holds the result of Match().
	vector<TagVal>	mMatches;

	// Methods.
	TagInt	GetTypeNum( const char* pTypeName );
	TagInt	GetFileNum( const char* pFileName );
	bool	GetTypeName( string& out, TagInt num );
	bool	GetFileName( string& out, TagInt num );
	int		TagCnt( void );
	bool	Insert( const char* pTagName, const char* pTagType, const char* pTagFile, int tagLine );
	bool	Free( const char* pFname );
	bool	IfTag( const char* pTagName );
	bool	IfTagRef( const char* pTagName, const char* pFileName, int line );
	PyObject* FindTag( const char* pTagName, bool caseSens, bool partial, bool regex );
    PyObject* PythonListFromTags( const vector<TagVal>& tags );
	int		Match( const char* pSoughtTagName, bool caseSens, bool partial=false, bool regex=false );
	int		MatchRegex( const char* pSoughtTagName, bool caseSens );
    int     GetModuleTags( const char* pModuleName );
};
static Tags sTags;

/*****************************************************************************
 * Return enumerations of tag types and filenames.
 * This saves memory by storing short ints for these instead of strings.
 *****************************************************************************/
// If number hasn't been seen before, insert it into a map and a reverse-map.
#define GET_NUM( CSTR, MAP_STR_NUM_TYPE, MAP_STR_NUM, MAP_NUM_STR, NUM_VAR )	\
{																				\
	string str(CSTR);															\
	MAP_STR_NUM_TYPE::iterator itr = MAP_STR_NUM.find(str);						\
    if ( itr != MAP_STR_NUM.end() )												\
		return itr->second;														\
	else																		\
	{																			\
		++NUM_VAR;																\
		MAP_STR_NUM.insert( make_pair<string,Tags::TagInt>(str, NUM_VAR) );		\
		MAP_NUM_STR.insert( make_pair<Tags::TagInt,string>(NUM_VAR, str) );		\
		return NUM_VAR;															\
	}																			\
}
Tags::TagInt
Tags::GetTypeNum( const char* pTypeName )
{
	GET_NUM( pTypeName, TypeStrNumMap, mTypeStrNumMap, mTypeNumStrMap, mTypeNum )
} 
Tags::TagInt
Tags::GetFileNum( const char* pFileName )
{
	GET_NUM( pFileName, FileStrNumMap, mFileStrNumMap, mFileNumStrMap, mFileNum )
} 

/*****************************************************************************
 * Return the string that corresponds to an enumeration.
 *****************************************************************************/
#define GET_NAME( NUM, MAP_NUM_STR_TYPE, MAP_NUM_STR )							\
{																				\
	MAP_NUM_STR_TYPE::iterator itr = MAP_NUM_STR.find( NUM );					\
	if ( itr != MAP_NUM_STR.end() )												\
	{																			\
		out = itr->second;														\
		return true;															\
	}																			\
	else																		\
	{																			\
		out = "";																\
		return false;															\
	}																			\
}
bool
Tags::GetTypeName( string& out, Tags::TagInt num )
{
	GET_NAME( num, TypeNumStrMap, mTypeNumStrMap )
}
bool
Tags::GetFileName( string& out, Tags::TagInt num )
{
	GET_NAME( num, FileNumStrMap, mFileNumStrMap )
}

/*****************************************************************************
 * Called by Exuberant ctags when it has created a tag for one src file.
 *****************************************************************************/
static void
TagCallback( const Extag* tag )
{
	sTags.Insert( tag->nameStr, tag->typeStr, tag->fileStr, tag->lineNum );
    return;
}

/*****************************************************************************
 * Interface to Python: Tag symbols in a src file.
 * Defers to Exuberant ctags (its src code is integrated with pypersrc).
 *****************************************************************************/
PyObject*
TagFile( PyObject* nil, PyObject* args )
{
	// Get args passed by Python.
	const char*	filename;
    if ( ! PyArg_ParseTuple( args, "s", &filename ) )
	{
		cerr << "ERROR: TagFile() failed \n";
		return NULL;
	}

    // Pass src filename to Exuberant ctags.
    // It will callback for each tag.
    int res = CtagsTagFile( filename, TagCallback );
    return Py_BuildValue( "i", res );
}

/*****************************************************************************
 * Interface to Python: Return amount of tags.
 *****************************************************************************/
PyObject*
TagCnt( PyObject* nil, PyObject* args )
{
	// Call C++ function, return result to Python.
	int res = sTags.TagCnt();
    return Py_BuildValue( "i", res );
}
int
Tags::TagCnt( void )
{
	return mTagMap.size();
}

/*****************************************************************************
 * Interface to Python: Insert a tag into C++ multimap.
 *****************************************************************************/
PyObject*
InsertTag( PyObject* nil, PyObject* args )
{
	// Get args passed by Python.
	const char* pTagName;
	const char* pTagType;
	int			tagLine;
	const char*	pTagFile;
    if ( ! PyArg_ParseTuple( args, "ssis", &pTagName, &pTagType, &tagLine, &pTagFile ) )
	{
		cerr << "ERROR: InsertTag() failed \n";
		return NULL;
	}

	// Call C++ function, return result to Python.
	int res = sTags.Insert( pTagName, pTagType, pTagFile, tagLine );
    return Py_BuildValue( "i", res );
}
bool
Tags::Insert( const char* pTagName, const char* pTagType, const char* pTagFile, int tagLine )
{
	// Insert tag into map.
	// To conserve memory, store a short int in lieue of a string
	// (the short int is mapped to the string).
    // Convert key to lower-case in order to support case-insensitive matching.
    // The unconverted tag name is stored in the value for case-sensitive matching.
	TagInt typeNum = GetTypeNum( pTagType );
	TagInt fileNum = GetFileNum( pTagFile );
    string key; StringLowercase( pTagName, key );
    string tagName(pTagName);
//	TagVal val( string(pTagName), typeNum, fileNum, tagLine );  // g++ 2.9 won't compile this
	TagVal val( tagName, typeNum, fileNum, tagLine );           // value holds tag name's original case
	mTagMap.insert( make_pair<string,TagVal>(key, val) );
	return true;
}

/*****************************************************************************
 * Interface to Python: Free tags belonging to a source file.
 *****************************************************************************/
PyObject*
FreeTags( PyObject* nil, PyObject* args )
{
	// Get args passed by Python.
	const char* pFname;
    if ( ! PyArg_ParseTuple( args, "s", &pFname ) )
	{
		cerr << "ERROR: FreeTags() failed \n";
		return NULL;
	}

	// Call C++ function, return result to Python.
	int res = sTags.Free( pFname );
    return Py_BuildValue( "i", res );
}
bool
Tags::Free( const char* pFname )
{
	string file1(pFname), file2;
	TagMap::iterator itr, itr2;
	for ( itr = mTagMap.begin(); itr != mTagMap.end(); )
	{
		GetFileName( file2, itr->second.mFile );
		if ( file1 == file2 )
		{
			itr2 = itr; ++itr2;		// don't increment an erased iterator!
			mTagMap.erase( itr );
			itr = itr2;
		}
		else
			++itr;
	}
	return true;
}

/*****************************************************************************
 * Interface to Python: Return true if this is a tag.
 *****************************************************************************/
PyObject*
IfTag( PyObject* nil, PyObject* args )
{
	// Get args passed by Python.
	const char* pTagName;
    if ( ! PyArg_ParseTuple( args, "s", &pTagName ) )
	{
		cerr << "ERROR: IfTag() failed \n";
		return NULL;
	}

	int res = sTags.IfTag( pTagName );
    return Py_BuildValue( "i", res );
}
bool
Tags::IfTag( const char* pTagName )
{
	return Match( pTagName, true );	// true if one or more matches found
}

/*****************************************************************************
 * Interface to Python: Return true if this is a reference to a tag.
 * For example, a reference to a tag should be drawn as a hyperlink.
 * But a tag should be drawn normally on the line where it is defined.
 *****************************************************************************/
PyObject*
IfTagRef( PyObject* nil, PyObject* args )
{
	// Get args passed by Python.
	const char* pTagName;
	const char* pFileName;
	int         line;
    if ( ! PyArg_ParseTuple( args, "ssi", &pTagName, &pFileName, &line ) )
	{
		cerr << "ERROR: IfTagRef() failed \n";
		return NULL;
	}
	int res = sTags.IfTagRef( pTagName, pFileName, line );
    return Py_BuildValue( "i", res );
}
bool
Tags::IfTagRef( const char* pTagName, const char* pFileName, int line )
{
	int file = GetFileNum(pFileName);
	int matches = Match( pTagName, true );	// result in mMatches vector
	for ( int i = 0; i < matches; ++i )
	{
		if (   mMatches[i].mFile == file
            && mMatches[i].mLine == line )
			return false;		// this is the line where the tag is defined
	}							// (it's a tag definition, not a tag reference)
	return true;
}

/*****************************************************************************
 * Interface to Python: Return list of matching tags or None.
 *****************************************************************************/
PyObject*
FindTag( PyObject* nil, PyObject* args )
{
	// Get args passed by Python.
	const char* pTagName;
    int         caseSens;
	int			partial;
	int			regex;
    if ( ! PyArg_ParseTuple( args, "siii", &pTagName, &caseSens, &partial, &regex ) )
	{
		cerr << "ERROR: FindTag() failed \n";
		return NULL;
	}
	return sTags.FindTag( pTagName, caseSens, partial, regex );
}
PyObject*
Tags::FindTag( const char* pTagName, bool caseSens, bool partial, bool regex )
{
	int matches = Match( pTagName, caseSens, partial, regex );	// result in mMatches vector
	if ( matches )
    {
        return PythonListFromTags( mMatches );
    }
	else
    {
		Py_INCREF(Py_None);
	    return Py_None;
	}
}

/*****************************************************************************
 * Build a Python list from a C++ vector (mMatches in practice).
 * Multiple tags may share the same name.
 * @return A list of tuples: [ (name,type,file,line), ... ]
 * @pre mMatches holds the matching tags.
 *****************************************************************************/
PyObject*
Tags::PythonListFromTags( const vector<TagVal>& tags )
{
	PyObject* pyList = PyList_New( tags.size() );
	for ( unsigned int i = 0; i < tags.size(); ++i )
	{
		// Get tag elements.
		string typeName, fileName;
   		int line = tags[i].mLine;
		if (  ! GetTypeName( typeName, tags[i].mType )
		   || ! GetFileName( fileName, tags[i].mFile ) )
		{
			cerr << "FindTag() malfunctioned \n";
		    Py_DECREF(pyList);
		    Py_INCREF(Py_None);
			return Py_None;
        }
       	// Put tag elements into a Python tuple.
   		PyObject* pyTuple = Py_BuildValue( "(sssi)",
                                           tags[i].mName.c_str(),
                                           typeName.c_str(),
                                           fileName.c_str(),
                                           line );
   		// Append tuple to list.
   		PyList_SetItem( pyList, i, pyTuple );
    }
	return pyList;
}

/*****************************************************************************
 * Match tags by name.  Result is the vector mMatch.
 *****************************************************************************/
int
Tags::Match( const char* pSought, bool caseSens, bool partial, bool regex )
{
	if ( regex ) return MatchRegex( pSought, caseSens );		

	mMatches.clear();       // no matches yet
	TagMap::iterator itr;

	if ( partial )
	{
		// Partial match searches key-by-key (slow).
		// '*' can be passed to match all tags.
        // For case-insensitivity, tagMap keys were converted to lower-case (itr->first).
        // For case-sensitivity, TagValue.mName holds the original case (itr->second.mName).
        if ( pSought[0] == '*' )
        {
    		for ( itr = mTagMap.begin(); itr != mTagMap.end(); ++itr )
				mMatches.push_back( itr->second );      // found a match
        }
        else if ( ! caseSens )
        {
            string soughtLower; StringLowercase( pSought, soughtLower );
    		for ( itr = mTagMap.begin(); itr != mTagMap.end(); ++itr )
    			if ( itr->first.find( soughtLower ) != string::npos ) // tagMap keys are lowercase
    				mMatches.push_back( itr->second );	// found a match
        }
        else if ( caseSens )
        {
    		for ( itr = mTagMap.begin(); itr != mTagMap.end(); ++itr )   // TagValue.mName
                if ( itr->second.mName.find( pSought ) != string::npos ) // has original case
    				mMatches.push_back( itr->second );	// found a match
        }
	}
	else // ! partial (match entire tag name)
	{
        string soughtLower; StringLowercase( pSought, soughtLower );
		for ( itr  = mTagMap.lower_bound( soughtLower );
			  itr != mTagMap.upper_bound( soughtLower );
	          ++itr )
		{
			// This tag's name does match when ignoring case.
			// But is case to be regarded?
			// TagVal.mName holds the tag name's original case.
			if ( ! caseSens || itr->second.mName == pSought )
				mMatches.push_back( itr->second );      // found a match
		}
	}

	return mMatches.size();
}

/*****************************************************************************
 * Match tags by name using a regex pattern.  Result is the vector mMatch.
 *****************************************************************************/
int
Tags::MatchRegex( const char* pSoughtTagName, bool caseSens )
{
	mMatches.clear();								// no matches yet

	// Compile regex pattern.
	regex_t		re;
	regmatch_t	pmatch[1];
	const int	eflags = 0;
	const int	cflags = (caseSens ? 0 : REG_ICASE) | REG_EXTENDED;
	if ( regcomp( &re, pSoughtTagName, cflags ) )
		return 0;

	// Try to match the regex pattern against every tag's name.
	TagMap::iterator itr;
	for ( itr = mTagMap.begin(); itr != mTagMap.end(); ++itr )
	{
		if ( regexec( &re, itr->second.mName.c_str(), 1, pmatch, eflags ) == 0 )
			mMatches.push_back( itr->second );		// found a match
	}

	return mMatches.size();
}

/*****************************************************************************
 * Interface to Python: Return list of matching tags or None.
 *****************************************************************************/
PyObject*
GetModuleTags( PyObject* nil, PyObject* args )
{
	// Get args passed by Python.
	const char* pModuleName;
    if ( ! PyArg_ParseTuple( args, "s", &pModuleName ) )
	{
		cerr << "ERROR: GetModuleTags() failed \n";
		return NULL;
	}

    // Find all tags in the module, put them into a C++ vector,
    // and convert the vector into a Python list.
    if ( sTags.GetModuleTags( pModuleName ) )  // result in mMatches
    {
        return sTags.PythonListFromTags( sTags.mMatches );
    }
	else
    {
		Py_INCREF(Py_None);
	    return Py_None;
	}
}
// Get all tags in one module (source file).  Result is the vector mMatch.
int
Tags::GetModuleTags( const char* pModuleName )
{
	mMatches.clear();                               // no matches yet
	TagMap::iterator itr;
    int fileNum = GetFileNum( pModuleName );        // convert file name to file number
	for ( itr = mTagMap.begin(); itr != mTagMap.end(); ++itr )
    {
        if ( itr->second.mFile == fileNum )         // is tag in the sought module/file?
        	mMatches.push_back( itr->second );      // found a match
    }

	return mMatches.size();
}
