/*****************************************************************************
 * $Id: pyext_Utils.cc,v 1.1 2006-02-25 05:39:54 devel Exp $
 * @file
 * @author	Jim E. Brooks  http://www.jimbrooks.org
 * @brief	Utility functions implemented in C++ for Python.
 * @remarks Python extension functions have the form:
 *          PyExt(PyObject* self, PyObject* args)
 *          I write PyExt(nil,args) because in fact self will be NULL
 *          (except for the esoteric "Python extension types").
 *          A few of my extensions need self which my Python code passes as
 *          the first arg which appears on the C++ side as the first element
 *          in the tuple passed as "args" to PyExt(self,args).
 *//*
 * LEGAL:	COPYRIGHT (C) JIM E. BROOKS 2006
 *          THIS SOURCE CODE IS RELEASED UNDER THE TERMS
 *          OF THE GNU GENERAL PUBLIC LICENSE (GPL).
 *****************************************************************************/

#include <Python.h>
#include "common.hh"

// Table of C++ function that act as Python methods (Python extensions).
PyMethodDef exportMethodsUtils[] =
{
    { "Utils_Find",			Utils_Find,			METH_VARARGS, "Search a string for substring." },
    { "Utils_FindLines",	Utils_FindLines,	METH_VARARGS, "Return list of line numbers that match." },
	{ NULL, NULL, 0, NULL }
};

/*****************************************************************************
 * Interface to Python: Search a string for a substring.
 * Like string.find(), if found, returns positive index of match, else -1.
 *****************************************************************************/
PyObject*
Utils_Find( PyObject* nil, PyObject* args )
{
	const char* haystack;
	const char* needle;
	int		    caseSens;
    if ( PyArg_ParseTuple( args, "ssi", &haystack, &needle, &caseSens ) )
	{
        // Assume no match.
        int i = -1;
		int line = 1;	// Find() requires initializing this as 1
        const char* match = Find( haystack, needle, caseSens, line );
        if ( match != NULL )
		{
            i = match - haystack;    // index of needle inside of haystack
		}

        // Return index of match.
        return Py_BuildValue( "i", i );
	}
	PY_RETURN_BAD_OBJECT_ERROR
}

/*****************************************************************************
 * Interface to Python: Returns a Python list of line numbers that match.
 * Returns [ lineNum, lineNum, ... ] or [] if no match.
 *****************************************************************************/
PyObject*
Utils_FindLines( PyObject* nil, PyObject* args )
{
	const char* haystack;
    int         haystackLen;
	const char* needle;
    int         needleLen;
	int			caseSens;
    if ( PyArg_ParseTuple( args, "s#s#i", &haystack, &haystackLen, &needle, &needleLen, &caseSens ) )
	{
        // First build a C++ vector of matching lines.
        vector<int> matchingLines;
        const char* haystack2 = haystack;
        const char* haystackEnd = haystack + haystackLen;
        int line = 1;    // Find() requires initializing this as 1
        while ( haystack2 < haystackEnd )
        {
            const char* match = Find( haystack2, needle, caseSens, line );
            if ( match )
            {
                matchingLines.push_back( line );
                // Ignore multiple matches on the same line.
                haystack2 = MovePastNewLine( match + 1 );
            }
            else break;
        }

        // Now that the number of matches is known (from the C++ vector),
        // create a Python list (Python C/API requires list size to be predetermined)
        // to hold the matches.
    	PyObject* pyList = PyList_New( matchingLines.size() );
        for ( uint i = 0; i < matchingLines.size(); ++i )
            PyList_SET_ITEM( pyList, i, Py_BuildValue("i",matchingLines[i]) );
        return pyList;
	}
	PY_RETURN_BAD_OBJECT_ERROR
}
