# $Id: Command.py,v 1.27 2006-03-26 03:03:32 devel Exp $
# pypersrc
# Author:   Jim Brooks  http://www.jimbrooks.org
# LEGAL:    COPYRIGHT (C) JIM E. BROOKS 2005
#           THIS SOURCE CODE IS RELEASED UNDER THE TERMS
#           OF THE GNU GENERAL PUBLIC LICENSE (GPL).
# Notes:    For consistency, commands with a file arg take it last.
# ==============================================================================

# See GuiText.py for mouse button behavior for text widgets.

helpText = \
"Keyboard/mouse:                                                                  \n" \
"---------------                                                                  \n" \
"button 1|>1   load into new|same page                                            \n" \
"F1            history of command results                                         \n" \
"F2            show results from last command                                     \n" \
"Ctrl-W        close page                                                         \n" \
"refresh       click this after you modify a source file                          \n" \
"up|down       previous|next command in history (Command entry widget)            \n" \
"                                                                                 \n" \
"Commands (basic):                                                                \n" \
"-----------------                                                                \n" \
"h[elp]|?                         help screen                                     \n" \
"n                                goto line                                       \n" \
"b[cr] <file>                     browse source file                              \n" \
"f[cr] <text> [[cr] [~]<file>]..  find text in source file(s)                     \n" \
"g[cr] <text>                     find text in all source files                   \n" \
"t[cre] <tag>|* [type]|* [file]   find a source code tag                          \n" \
"                                                                                 \n" \
"Matching:                                                                        \n" \
"[c] case-sensitive, [r] regex, [e] exact, [*] match any, [~] negative            \n" \
"For filename matching, globbing is implied if pattern contains a '*?[]' char.    \n" \
"                                                                                 \n" \
"Examples:                                                                        \n" \
"b debug              list src files whose names match 'debug'                    \n" \
"b *.c                lists C src files                                           \n" \
"t *                  lists all tags                                              \n" \
"t * func             lists tags of all functions                                 \n" \
"t Get func           lists tags of all functions named Get (partial matching)    \n" \
"t Get func *.c       lists tags of all functions in all C src files              \n" \
"t Get * *.c          lists all tags partially matching Get in all C src files    \n" \
"t * * *.c            lists all tags in all C src files                           \n" \
"f error              search for 'error' in current file                          \n" \
"f error *            search for 'error' in all files                             \n" \
"g error              same as 'f error *'                                         \n" \
"fc Error             case-sensitive search for 'Error'                           \n" \
"fc Error *.c         likewise but only search in C files                         \n" \
"fc Error c *.c       likewise but file matching is case-sensitive, excludes *.C  \n" \
'f Error *.cc *.py    include C++ and Python file in global search                \n' \
'f Error ~*.cc ~*.py  exclude C++ and Python files in global search               \n' \
"fr return.*false     search for regular expression                               \n" \
'f "Hello World"      surround with quotes a pattern that is multiple words       \n' \
'f ""Hello World""    repeat leading or trailing quote to include it in pattern   \n' \
"                                                                                 \n" \
"Commands (advanced):                                                             \n" \
"--------------------                                                             \n" \
"class                show class diagram (TODO)                                   \n" \
"trace                show chains of function calls (TODO)                        \n" \
"config               edit configuration (TODO)                                   \n" \
"                                                                                 \n" \
"Commands (misc):                                                                 \n" \
"----------------                                                                 \n" \
"hilite no|lo|hi      level of syntax highlighting (affects loading speed)        \n" \
"                                                                                 \n" \
"Commands (debug):                                                                \n" \
"-----------------                                                                \n" \
"log                  show log                                                    \n" \
"log [c]lear          clear log                                                   \n" \
"log save <file>      save log to file                                            \n" \
"                                                                                 \n" \
"Command-line arguments:                                                          \n" \
"-----------------------                                                          \n" \
"[dir..] [files..]    src files and/or directories to browse                      \n" \
"-x[c] <glob>         exclude files [c: case-sensitive]                           \n" \
"                                                                                 \n" \
"Website:                                                                         \n" \
"http://www.jimbrooks.org/hypersrc/  ('hypersrc' isn't a typo)                    \n"

#-------------------------------------------------------------------------------
# Class Command.
#-------------------------------------------------------------------------------

class Command(object):

    # Internal commands (as if the user entered them in the Command entry).
    # Internal and user commands are stored in the command history.
    INTERNAL_COMMAND_CLICK_TAG    = "[click_tag]"
    INTERNAL_COMMAND_SPLASH       = "[splash]"

    def __init__( s, argList ):
        s.commandList    = []          # list of user commands (for command history)
        s.commandListIdx = 0           # command recalled in command history

    # INTERFACE
    def Recall( s, next=False, prev=False ):
        """ Recall the next or previous command.  Returns empty string if no commands. """
        if len(s.commandList) == 0:
            return ""
        else:
            idx = s.commandListIdx
            # Use negative indexing.
            if next:
                idx += 1
                if idx >= 0:
                    idx = -1
            else:
                idx -= 1
                if idx < -len(s.commandList):
                    idx = -len(s.commandList)
        command = s.commandList[idx]    # shouldn't throw
        s.commandListIdx = idx          # commit index
        return command

    # INTERFACE
    def Do( s, command, resultsTextWid=None ):
        """ Process command given by user (or internal command). """
        try:
            if not command:
                return

            # s.resultsTextWid is to be used henceforth and in subroutines.
            if resultsTextWid is None:
                s.resultsTextWid = st.gui.GetResultsTextWid()
            else:
                s.resultsTextWid = resultsTextWid
            
            # Remember this command.
            # Even a bad command is remembered so that
            # the user has a chance to the recall and fix it.
            s.commandList.append( command )
            s.commandListIdx = 0     # this resets command history recall

            # Only commandWords[0] should always be present.
            # Accessing commandWords[>0] could cause an exception!
            commandWords = SplitWords( command )    # handles quoted words

            # Goto line number?
            if commandWords[0].isdigit():
                line = int(commandWords[0])  # convert string from user to int
                s.__GotoLine( line )
            # Browse source file?
            elif commandWords[0] in [ "b", "br", "bc" ]:
                s.__BrowseSrcFile( command, commandWords )
            # Find text?
            elif commandWords[0][0] in "fg":
                s.__FindText( command, commandWords )
            # Find tag?
            elif commandWords[0] in [ "t", "tc", "tr", "tcr", "trc", "te" ]:
                s.__FindTag( command, commandWords )
            # Syntax hiliting/highlighting.
            elif commandWords[0] == "hilite":
                hiliteLevel = commandWords[1]
                LexBase.SetHilite( hiliteLevel )  # call Python static method
                st.gui.FlushSrcFiles()
                st.gui.ReloadSrcFiles()
            # Help?
            elif commandWords[0] in [ "h", "help", "?" ]:
                tokenList = [ (eTokenText,helpText) ]
                contents = ContentsResults( tokenList=tokenList, fname="[results: help]", command=command )
                s.resultsTextWid.LoadContents( contents )
            # Log command?
            elif commandWords[0] == "log":
                s.__Log( commandWords )
            # (internal command)
            elif commandWords[0] == s.INTERNAL_COMMAND_CLICK_TAG:
                # Re-click tag for refresh.
                del s.commandList[-1]    # don't store internal commands in history
                clickedTagName = commandWords[1]
                clickedFname   = commandWords[2]
                st.gui.TagClick( clickedTagName, clickedFname )
            #  (internal command)
            elif commandWords[0] == s.INTERNAL_COMMAND_SPLASH:
                # Splash screen shows help instructions.
                del s.commandList[-1]    # don't store internal commands in history
                pythonVersion = "Python " + sys.version
                text = "pypersrc " + st.version + "\n\n" + pythonVersion + "\n\n" + helpText
                fname = "[results: splash]"
                s.__ShowTempResults( fname, text )
            # Easter egg?
            elif commandWords[0] in [ "easter", "egg", "easteregg", "underdog" ]:
                Underdog()
                s.Do( "log" )
            # Unknown command.
            else:
                st.gui.statusbar.Print( "Unknown or unimplemented command.  Type 'h' for help." )
                return

        except:
            PrintException()

    # PRIVATE
    def __ShowTempResults( s, fname, text ):
        """ Show temp results that won't be saved in history. """
        tokenList = [ (eTokenText,text) ]
        contents = ContentsResults( tokenList=tokenList, fname=fname, command=None )
        contents.saveInHistory = False   # forget
        s.resultsTextWid.LoadContents( contents )

    # PRIVATE
    def __Log( s, commandWords ):
        """ Execute a log command. """
        # commandWords[>0] could be absent and access will cause exception.
        
        # Clear log?
        try:
            if commandWords[1] in [ "clear", "c" ]:
                st.log.Clear()
                s.resultsTextWid.LoadLog()
                return
        except:
            pass

        # Save log to file?
        try:
            if commandWords[0] == "log" and commandWords[1] == "save":
                if len(commandWords) != 3:
                    st.gui.statusbar.Print( "One argument is required." )
                else:
                    try:
                        fname = commandWords[2]
                        fd = os.open( fname, os.O_CREAT|os.O_TRUNC|os.O_WRONLY, 0600 )
                        os.write( fd, st.log.Read() )
                        os.close( fd )
                        text = "Saved log into file " + AbsPath( fname )
                        st.gui.statusbar.Print( text )
                        st.log.Write( text )
                    except:
                        PrintException()
                        text = "Failed to write log to file '" + fname + "'"
                        st.gui.statusbar.Print( text )
                        st.log.Write( text )
                return
        except:
            pass

        # Show log in text widget.
        s.resultsTextWid.LoadLog()
        
    # PRIVATE
    def __GotoLine( s, line ):
        """ Goto line number in current text widget. """
        textWid = st.gui.notebookSrcFile.GetCurrentTextWid()
        if textWid:
            st.log.Write( "Command: goto line " + str(line) + " in " + textWid.GetFname() )
            textWid.GotoLine( line=line, hiliteLine=True )
        
    # PRIVATE
    def __FindText( s, command, commandWords ):
        """ Find text in all source files. """

        # Determine scope of search.
        # Use logical AND with length to ensure element index is valid.
        # Treat "g <text> <file>" as constrained search for convenience.
        searchOne  = (command[0][0] == "f") and (len(commandWords) == 2)
        searchAll  = (command[0][0] == "g" and len(commandWords) == 2) \
                  or (command[0][0] == "f" and len(commandWords) == 3 and commandWords[2] == "*")
        searchSome = not searchAll and len(commandWords) > 2
        if int(searchOne) + int(searchAll) + int(searchSome) != 1:
            raise Oops( "FindText: confused about scope of search!" )

        # Gather srcFileList according to scope of search.
        if searchOne:
            (srcFileList,errorText) = s.__FindTextOne( command, commandWords )
        elif searchAll:
            (srcFileList,errorText) = s.__FindTextAll( command, commandWords )
        elif searchSome:
            (srcFileList,errorText) = s.__FindTextSome( command, commandWords )
        else:
            raise Oops( "FindText: unhandled case!" )

        # srcFileList now should have list of files to search.
        if srcFileList:
            # Search in a second thread as searching can be slow.
            pattern = commandWords[1]
            modifiers = commandWords[0][1:]
            finderText = s.__Finder( pattern, modifiers, findFile=False )
            st.threads.Spawn( ThreadFindText, command, finderText, srcFileList )
        elif errorText:
            st.gui.statusbar.Print( errorText )

    # PRIVATE
    def __FindTextOne( s, command, commandWords ):
        """ Search one file.  Returns (srcFileList,errorText). """
        
        # But is there yet a current text widget (notebook empty)?
        textWid = st.gui.notebookSrcFile.GetCurrentTextWid()
        if textWid:
            fname = textWid.contents.fname
            srcFileList = [ fname ]
            errorText = ""
        else:
            srcFileList = []
            errorText = "Text widget is empty, nothing to find locally."
        return (srcFileList,errorText)

    # PRIVATE
    def __FindTextAll( s, command, commandWords ):
        """ Search all files.  Returns (srcFileList,errorText). """
        if commandWords[0] == "g" and len(commandWords) != 2:
            srcFileList = []
            errorText = "g command: requires one arg."
        else:
            srcFileList = st.gather.GetSrcFileList()
            errorText = ""
        return (srcFileList,errorText)

    # PRIVATE
    def __FindTextSome( s, command, commandWords ):
        """ Search in a constrained subset of src files:
            f <text> [cr] [~]<file>...
            Returns (srcFileList,errorText). """
        
        # Build list of Finder objects (one for each file pattern the user specified).
        finderFileList = []
        modifiersFile = ""     # modifiers for file pattern, in separate word ('~' prefixes the pattern)
        anyNegative = False    # if negative match was specified one or more times
        i = 2                  # index of first file pattern (one or more)
        while i < len(commandWords):
            if len(commandWords[i]) <= 2 and IfHaveEachChar( commandWords[i], "cr" ):
                # This is a pattern modifier (case-sensitive and/or regex).
                modifiersFile = commandWords[i]
            else:
                # This is a pattern for a file.  It could have a negative match prefix ('~').
                patternFile = commandWords[i]
                negative = (patternFile[0] == "~")    # negative match?
                if negative:
                    anyNegative = True
                    patternFile = patternFile[1:]    # chop '~'
                finderFile = s.__Finder( patternFile, modifiersFile, negative=negative, findFile=True )
                finderFileList.append( finderFile )
            i += 1
        if not finderFileList:
            st.gui.statusbar.Print( "f command: unrecognized args, try quoting pattern, or type 'help'." )
            return
        
        # Build list of matching src files.
        srcFileList = []
        for fname in st.gather.GetSrcFileList():
            # Test if this src file matches pattern(s).
            if not anyNegative:
                # Multiple non-negative matches imply logical OR.
                # Eg, "f text *.cc *.py" means C++ and Python files.
                for finderFile in finderFileList:
                    if finderFile.IfMatch( fname ):
                        srcFileList.append( fname )
                        break
            else:
                # Multiple negative matches imply logical AND.
                # Eg "f text ~*.cc ~*.py" means all files excluding C++ and Python.
                mismatch = False
                for finderFile in finderFileList:
                    # This can be confusing: IfMatch() returns True if fname does NOT match
                    # (negative matching is being used), then "not" inverts.
                    if not finderFile.IfMatch( fname ):
                        mismatch = True
                        break
                if not mismatch:    # if all negative matches were true (logical AND)
                    srcFileList.append( fname )
                        
        if srcFileList:
            errorText = ""
        else:
            errorText = "The file pattern doesn't match any filenames (nothing to search)."
        return (srcFileList,errorText)

    # PRIVATE
    def __FindTag( s, command, commandWords ):
        """ Find a source code tag. """
        # "t <name> <type> <file>"  constrain by name, type, file
        # "t <name> * <file>"       constrain by name, file (* means match any and a syntactic placeholder)
        # "t * * <file>"            constrain by file
        # For consistency with the f/g/b commands, the fileglob is the t command's last arg.

        try:
            caseSens = ("c" in commandWords[0])
            regex    = ("r" in commandWords[0])
            exact    = ("e" in commandWords[0])

            tokenList = []
            tagName = commandWords[1]

            # Optionally constrain the search by tag type.
            try:
                tagType = commandWords[2]    # user may omit
                if tagType == "*":           # '*' means any type, used as syntactic placeholder
                    tagType = None           # to constrain by fileglob only
            except:
                tagType = None

            # Optionally constrain the search by tag file (fileglob).
            try:
                tagFile = commandWords[3]    # user may omit
            except:
                tagFile = None

            # To tell user if search was constrained.
            constraints = ""
            if caseSens:
                constraints = AddIntoParentheses( constraints, "case-sensitive" )
            if tagType:
                constraints = AddIntoParentheses( constraints, "type:" + tagType )
            if tagFile:
                constraints = AddIntoParentheses( constraints, "file:" + tagFile )

            # Find tags.
            if exact:
                tagList = st.tags.FindTag( tagName, tagType, tagFile, caseSens=1, partial=0, regex=0 )
            elif regex:
                tagList = st.tags.FindTag( tagName, tagType, tagFile, caseSens,   partial=0, regex=1 )
            else:
                tagList = st.tags.FindTag( tagName, tagType, tagFile, caseSens,   partial=1, regex=0 )

            # Prepare to load text widget with hyperlinks to matching tags.
            if tagList:
                tokenList.insert( 0, (eTokenText,"Found tags matching `" + tagName + "`" + constraints + ":\n\n") )
                for tag in tagList:
                    # Build a phony tag from the real tag whose name is the elements
                    # of the real tag (name,type,file,line).  The reason is that
                    # phony tags, unlike real tags, when put into a token list,
                    # are shown as hyperlinks whose color changes if visited.
                    phonyTagName = tag[0] + " " + tag[1] + " " + tag[2] + " " + str(tag[3])
                    phonyTagFile = tag[2]
                    phonyTagLine = tag[3]
                    phonyTag = st.tags.PhonyTag( phonyTagName, phonyTagFile, phonyTagLine )
                    tokenList.append( (eTokenTag,phonyTag) )
            else:
                tokenList.insert( 0, (eTokenText,"No tags match `" + tagName + "`" + constraints + ".\n\n") )
                # Does this look like a regex pattern but the user didn't specify the "x" suffix?
                if tagName != "*" and (tagName.find( "*" ) != -1 or tagName.find( "." ) != -1 or tagName.find( "[" ) != -1):
                    if not commandWords[0] in [ "tx", "txc", "tcx" ]:
                        tokenList.append( (eTokenText,"What you're searching for looks like a regex pattern.\n") )
                        tokenList.append( (eTokenText,"Use 'tx' rather than the 't' command to specify a regex pattern.\n") )
                        tokenList.append( (eTokenText,"Just press Up in the Command entry and append 'x' to 't'.\n") )

            # Load text widget with hyperlinks to matching tags.
            fname = "[results: " + command + "]"
            contents = ContentsResults( tokenList=tokenList, fname=fname, command=command )
            s.resultsTextWid.LoadContents( contents )
        except:
            PrintException()

    # PRIVATE
    def __BrowseSrcFile( s, command, commandWords ):
        """ Find a source file to browse. """
        try:
            if len(commandWords) != 2:
                st.gui.statusbar.Print( "One argument is required." )
                return
                    
            sought = commandWords[1]
            finder = s.__Finder( sought, commandWords[0][1:], findFile=True )
            fnameResults = "[results: " + command + "]"
            
            # Search listbox for rows that match the sought src file.
            # Build a list of matches (by index).
            rowTextList = st.gui.GetFileList()
            rowIdx = 0
            matchIdxList = []
            for rowText in rowTextList:
                if finder.IfMatch( rowText ):
                    matchIdxList.append( rowIdx )
                rowIdx += 1

            # Found one match?
            if len(matchIdxList) == 1:
                # Auto-select the matching src file.
                st.gui.DeselectFileList( matchIdxList[0] )
                fname = rowTextList[matchIdxList[0]]
                st.gui.notebookSrcFile.LoadSrcFile( fname )
            # Found multiple matches?
            elif len(matchIdxList) > 1:
                # Build results (hyperlinks to matching src files) in text widget
                # to let the user pick one.
                tokenList = []
                tokenList.insert( 0, (eTokenText,"Found multiple source files:\n\n") )
                for idx in matchIdxList:
                    fname = rowTextList[idx]
                    line = 0
                    phonyTag = st.tags.PhonyTag( fname, fname, line )
                    tokenList.append( (eTokenTag,phonyTag) )
                contents = ContentsResults( tokenList=tokenList, fname=fnameResults, command=command )
                s.resultsTextWid.LoadContents( contents )
            # Found no matches?
            else:
                tokenList = []
                text = "No source files match `" + sought + "`"
                tokenList.insert( 0, (eTokenText,text) )
                contents = ContentsResults( tokenList=tokenList, fname=fnameResults, command=command )
                s.resultsTextWid.LoadContents( contents )
        except:
            PrintException()

    # PRIVATE
    def __Finder( s, pattern, modifiers, negative=False, findFile=False ):
        """ Helper to return a Finder object according to match modifier chars. """
        # The user can enter "cre" as match modifier suffixes (in stand-alone word).
        # "~" for a negative match is special as it prefixes the pattern, not in a stand-alone modifier word.
        caseSens = "c" in modifiers    # case-sensitive matching
        regex    = "r" in modifiers    # regex matching
        exact    = "e" in modifiers    # exact matching
        glob     = False
        # File-globbing is implied by the chars in the pattern, a file being sought,
        # and other higher-precedence modifiers being false.
        if findFile and not (exact or regex):
            for c in pattern:
                if c in "*?[]":
                    glob = True
                    break
        try:
            return Finder( pattern, caseSens=caseSens, regex=regex, exact=exact, glob=glob, negative=negative )
        except:
            if regex:
                st.gui.statusbar.Print( "Bad regex.  Try quoting it (?)." )
            raise
