# $Id: Gather.py,v 1.38 2006-03-26 03:03:32 devel Exp $
# pypersrc
# Author:   Jim Brooks  http://www.jimbrooks.org
# LEGAL:    COPYRIGHT (C) JIM E. BROOKS 2005
#           THIS SOURCE CODE IS RELEASED UNDER THE TERMS
#           OF THE GNU GENERAL PUBLIC LICENSE (GPL).
# ==============================================================================

#-------------------------------------------------------------------------------
# Class Gather.
#-------------------------------------------------------------------------------

class Gather(MutexBase):
    """ Gathers a list of source files to be browsed. """

    def __init__( s, argList ):
        MutexBase.__init__( s )
        
        s.ProcessArgsPre( argList )
        
        # This defines the list of source files to browse.
        # A source file's modification time is remembered
        # to skip re-tagging it by refresh if it wasn't modified.
        s.srcFileDict       = {}                 # value is file modification time
        s.srcFileListSorted = []                 # should be equivalent to sorting keys of s.srcFileDict
        s.savedArgList      = []
        s.cwd               = os.getcwd()
        s.excludeFinderList = []                 # list of Finder objects that exclude files/dirs
        
        s.ProcessArgsPost( argList )

    # INTERFACE
    def ProcessArgsPre( s, argList ):
        """ Assigns None to elements in argList that were processed. """
        return True

    # INTERFACE
    def ProcessArgsPost( s, argList ):
        """ Assigns None to elements in argList that were processed. """
        try:
            i = 0
            while i < len(argList):
                arg = argList[i]
                if arg is None:
                    i += 1
                elif arg == "-x" or arg == "-xc":
                    # -x[c] <glob> excludes files/dirs.  [c] means case-sensitive.
                    # -x should be followed by one glob (not multiple).
                    # Design note: "-x glob1 glob2" would open the door to misinterpretation.
                    # Eg "pypersrc -x glob ." means exclude glob and browse CWD (not glob2='.').
                    argList[i] = None
                    i += 1
                    # The next arg should be one glob pattern.
                    if i < len(argList):
                        pattern = argList[i]
                        caseSens = (arg == "-xc")
                        finder = Finder( pattern=pattern, glob=True, caseSens=caseSens )
                        s.excludeFinderList.append( finder )
                        argList[i] = None
                        i += 1
                    else:
                        PrintWarning( "Gather: -x <glob> is missing glob arg." )
                else:
                    i += 1
        except:
            PrintException()
        return True

    # INTERFACE
    def Gather( s, argList ):
        """ Gathers/updates a list of source files to be browsed.  List is sorted by fname. """
        # - Updates s.srcFileDict,s.srcFileListSorted.
        #   This is a dictionary of filename,modificationTime pairs.
        #   The idea is to only re-tag files that were modified.
        # - Returns a list of sources files that need to be tagged.
        #   This excludes unmodified files.
        # - This accepts cmd-line args but is passed file/directory names
        #   by the GUI code (by a file dialog).
        return s.MutexCall( s.__Gather, argList )

    # PRIVATE
    def __Gather( s, argList ):
        # Temps used across methods.
        s.passCnt = 1    # information for user, changes each pass
        s.fileCnt = 0    # information for user, changes each pass (this isn't the final src file cnt!)
        s.dirCnt  = 0    # information for user
        
        st.log.Write( "Gather: gathering source files in programming languages " + str(st.langList) )
        
        # Log file/dir args (not yet proven that they are).
        fileArgInfo = ""
        for arg in argList:
            if arg is not None:
                fileArgInfo = AddIntoParentheses( fileArgInfo, arg )
        st.log.Write( "Gather: file/dir args: " + fileArgInfo )

        s.savedArgList = argList[:]
        s.srcFileDict.clear()

        def GatherUpdateStatusbar( gather ):
            # -- executed by main thread --
            # Gather.passCnt/fileCnt are just ints, using a mutex isn't necessary.
            text = "Gathering[" + str(gather.passCnt) + "]: " + str(gather.fileCnt) + " files"
            st.gui.statusbar.Print( text, baton=True )

        # First pass.
        # Build fileList from args.
        # Append args which are filenames.
        # For args which are directories, recurse into them, append filenames.
        s.passCnt = 1    # changed each pass
        s.fileCnt = 0    # changed each pass
        fileList = []
        st.gui.statusbar.SetUpdateCallback( GatherUpdateStatusbar, s )   # pass this object as "s"
        for arg in argList:
            try:
                if arg is None:
                    continue
                elif arg[0] == "-":                 # if not a hypenated cmd-line arg
                    continue
                elif os.path.isdir( arg ):          # dir?
                    dir = AbsPath( arg, s.cwd )
                    s.__GatherDir( dir, fileList )  # appends to fileList
                elif os.path.isfile( arg ):         # file?
                    fileList.append( os.path.normpath(arg) )
                    s.fileCnt += 1
                else:
                    s.__WarnUnrecognizedArg( arg )
            except:
                PrintException()

        # Second pass with gathered filenames found in dirs.
        s.passCnt = 2    # changed each pass
        s.fileCnt = 0    # changed each pass
        try:
            for file in fileList:
                s.fileCnt += 1
                
                # Is this a source file to browse?
                if s.__IfSrcFileToBrowse( file ):
                    file = TransformPath( file )
                    # OPTIMIZE: Logging every file is a bottleneck!
                    #st.log.Write( "Gather: found file to browse: " + file )
                    #st.log.Write( "Gather: found file to browse: " + file, Log.VERBOSE )
                    # Store file's modification time.
                    s.srcFileDict[file] = FileModificationTime( file )
        except:
            PrintException()
        st.gui.statusbar.SetUpdateCallback( None )
        fileList = None   # superceded by s.srcFileDict

        if st.escKey:
            st.log.Write( "Gather: stopped by user" )
        else:
            st.log.Write( "Gather: gathered " + str(len(s.srcFileDict)) + " src files in " +str(s.dirCnt) + " dirs" )

        # Return sorted list of source files.
        st.gui.statusbar.Print( "Sorting gathered source files...", timeout=0 )
        s.srcFileListSorted = s.srcFileDict.keys()
        s.srcFileListSorted.sort()
        st.gui.statusbar.Clear()
        return s.srcFileListSorted

    # PRIVATE
    def __Regather( s ):
        s.Gather( s.savedArgList )

    # PRIVATE
    def __GatherDir( s, dirPathName, fileList ):
        """ Recursively gather source files in a directory. """
        # Modifies fileList in-place.

        s.dirCnt += 1
        prettyDirName = PrettyDirName(dirPathName)

        # OPTIMIZE: These are too slow.
        #st.log.Write( "Gather: recursing dir " + prettyDirName, Log.VERBOSE )
        #st.gui.statusbar.Print( "Gathering " + prettyDirName, timeout=0 )
        
        for entry in os.listdir( dirPathName ):
            # Use absolute pathnames only.  Mixing relative and absolute
            # may cause the problem of duplicate files.
            pathName = os.path.join( dirPathName, entry )
            pathName = os.path.join( s.cwd, pathName )
            if entry in st.ignoreDirList:
                st.log.Write( "Gather: ignoring dir " + PrettyDirName(pathName) )
            elif s.__IsExcluded(entry):
                # Exclude file or dir.
                st.log.Write( "Gather: excluding " + pathName )
            elif os.path.isdir( pathName ) and not st.escKey: # don't recurse if user pressed Esc
                s.__GatherDir( pathName, fileList )
            elif os.path.isfile( pathName ):
                fileList.append( os.path.normpath(pathName) )
                s.fileCnt += 1

    # INTERFACE
    def Refresh( s ):
        """ Refresh gathered source files. """
        # Returns (change=True/False, modifiedList, removedList, addedList, unchangedList)
        return s.MutexCall( s.__Refresh )

    # PRIVATE
    def __Refresh( s ):
        modifiedList  = []
        removedList   = []
        addedList     = []
        unchangedList = []
        
        # Re-gather source files, then examine any changes relative
        # to the previous gathering.  Re-gathering might add source files,
        # encounter modified ones, or omit removed ones.
        savedSrcFileDict = s.srcFileDict.copy()
        s.__Regather()    # updates s.srcFileDict

        # Determine which source files were modified, removed, or unchanged.
        for file in savedSrcFileDict.keys():
            if s.srcFileDict.has_key( file ):
                if s.srcFileDict[file] == savedSrcFileDict[file]:   # value is file modification time
                    unchangedList.append( file )
                else:
                    modifiedList.append( file )
            else:
                removedList.append( file )

        # Determine which source files were added.
        for file in s.srcFileDict.keys():
            if not savedSrcFileDict.has_key( file ):
                addedList.append( file )

        change = len(modifiedList) + len(removedList) + len(addedList) > 0
        return (change, modifiedList, removedList, addedList, unchangedList)

    # PRIVATE
    def __IfSrcFileToBrowse( s, fname ):
        """ Return true if this is a source file to browse according to its file extension. """
        try:
            ext  = FileExt( fname )       # get file extension
            lang = progLangExtDict[ext]   # if file extension is mapped to a known prog lang
            if lang in st.langList:       # if prog lang is supposed to be browsed
                return True
            else:
                return False
        except:
            return False

    # INTERFACE
    def IfSrcFile( s, fname ):
        """ Return true if this is a source file that was gathered. """
        try:
            s.mutex.acquire()
            retval = s.srcFileDict.has_key( fname )
            return retval
        finally:
            s.mutex.release()   # this executes always, even if try returns

    # INTERFACE
    def GetSrcFileList( s ):
        """ Return a list of source files that were gathered. """
        try:
            s.mutex.acquire()
            retval = s.srcFileListSorted[:]   # return copy to avoid thread conflicts
            return retval
        finally:
            s.mutex.release()

    # PRIVATE
    def __IsExcluded( s, fname ):
        """ Return true if file/dir was excluded by -x[c] cmd-line arg. """
        for finder in s.excludeFinderList:
            if finder.IfMatch( fname ):
                return True
        return False

    # PRIVATE
    def __WarnUnrecognizedArg( s, arg ):
        text = "Gather: unrecognized cmd-line arg '" + arg + "' which isn't a file/dir."
        # Show tip only once.
        try:
            if s.showedTipSubargsAreLimited:
                pass
        except:
            s.showedTipSubargsAreLimited = True
            text += "\nTip: Don't pass multiple sub-args on cmd-line: \n" \
                    "wrong: pypersrc -l c c++    \n" \
                    "right: pypersrc -l c -l c++ \n"
        PrintWarning( text )
