# $Id: Gui.py,v 1.115 2006-03-26 03:03:32 devel Exp $
# pypersrc
# Author:   Jim Brooks  http://www.jimbrooks.org
# LEGAL:    COPYRIGHT (C) JIM E. BROOKS 2005
#           THIS SOURCE CODE IS RELEASED UNDER THE TERMS
#           OF THE GNU GENERAL PUBLIC LICENSE (GPL).
# Notes:
# ==============================================================================

# ==============================================================================
# These are base classes that provide an interface independent of any
# GUI toolkit in order to facilitate possibly porting to a different one.
# And they summarize the Python code.
# ------------------------------------------------------------------------------
# Thread-safety:
# These are the only Gui methods that are thread-safe:
# - Statusbar.Print()
# - Statusbar.Clear()
# The initial thread is dedicated to the Gui.
# Other threads are restricted to calling only thread-safe Gui methods.
# ==============================================================================

#-------------------------------------------------------------------------------
# Class GuiBase.
#-------------------------------------------------------------------------------

class GuiBase(object):
    """ GUI base class. """

    def __init__( s, argList ):
        st.gui = s    # Gui components need st.gui during construction/bringup
        
        # --- may be overridden by cmd-line args ---
        s.geom            = "1000x700"
        if IfCygwin():
            s.fontName    = "courier"
            s.fontSize    = 10
        else:
            s.fontName    = None
            s.fontSize    = None            
        # --- may be overridden by cmd-line args ---
        
        s.ProcessArgsPre( argList )
        
        s.enabled         = False
        s.nestEnable      = Nest( nestFirst=False )  # to nest Gui.Enable(T/F)
        s.mutex           = threading.RLock()        # reentrant lock
        
        s.ProcessArgsPost( argList )
        
    def ProcessArgsPre( s, argList ):
        """ Assigns None to elements in argList that were processed. """
        try:
            i = 0
            while i < len(argList):
                arg = argList[i]
                if arg is None:
                    i += 1
                elif arg == "-geom":
                    s.geom = argList[i+1]
                    argList[i]   = None
                    argList[i+1] = None
                    i += 2
                elif arg in [ "-font", "-font-name", "-fontname" ]:
                    s.fontName = argList[i+1]
                    argList[i]   = None
                    argList[i+1] = None
                    i += 2
                elif arg in [ "-font-size", "-fontsize" ]:
                    s.fontSize = argList[i+1]
                    argList[i]   = None
                    argList[i+1] = None
                    i += 2
                else:
                    i += 1
        except:
            PrintException()
        return True

    def ProcessArgsPost( s, argList ):
        """ Assigns None to elements in argList that were processed. """
        return True

    def IfReady( s ):
        """ If GUI is ready/running (not to be confused with enabled). """
        raise Oops( "abstract method" )

    def IfEnabled( s ):
        """ If GUI is sensitive to input. """
        return s.enabled

    def Enable( s, enable=True ):
        """ Make GUI sensitive to input. """
        # Derived class should override and call this.
        s.enabled = enable

    def EnableWidget( s, widget, enable=True ):
        """ Enable widget (or disable if enable=False). """
        raise Oops( "abstract method" )
    
    def DisableWidget( s, widget ):
        """ Unconditionally disable widget.
            Returns old enable state of widget. """
        s.EnableWidget( widget, False )

    def EventLoop( s ):
        raise Oops( "abstract method" )
        
    def SetTitle( s, title ):
        """ Set text of title bar of main window. """
        raise Oops( "abstract method" )

    def SetFileList( s, fileList ):
        """ Set listbox with list of source filenames to browse. """
        raise Oops( "abstract method" )

    def GetFileList( s ):
        """ Return list of source filenames. """
        raise Oops( "abstract method" )

    def DeselectFileList( s, idx ):
        """ Deselect row(s) in the listbox of src files. """
        raise Oops( "abstract method" )

    def GetResultsTextWid( s ):
        """ Return the text widget used to hold results. """
        raise Oops( "abstract method" )
    
    def FlushSrcFiles( s, parseModes=None ):
        """ Flush all source files out of cache. """
        # Without args, flushes entire cache.
        # With parseModes=n, flushes src files filtered in specified parse modes.
        raise Oops( "abstract method" )

    def ReloadSrcFiles( s ):
        """ Reload those text widgets that are showing source code. """
        raise Oops( "abstract method" )
        
    def TagClick( s, clickedTagName, clickedFname, currentPage=False ):
        """ The user clicked a hyperlink in a source file.
            Also called for the [click tag] internal command (for refreshing). """
        raise Oops( "abstract method" )
        
    def ProcessPendingEvents( s ):
        """ Process any pending GUI events. """
        raise Oops( "abstract method" )

    def ScheduleIdleFunc( s, func, *args ):
        """ Schedule a function to be called (once) when GUI is idle. """
        raise Oops( "abstract method" )

    def ScheduleTimerFunc( s, milsec, func, *args ):
        """ Schedule a function to be called (once) in the next n milliseconds. """
        raise Oops( "abstract method" )

#-------------------------------------------------------------------------------
# Class StatusbarBase.
#-------------------------------------------------------------------------------

class StatusbarBase(object):

    def __init__( s, frame ):
        s.updateCallback     = None  # a client of Statusbar can optionally register a callback
        s.updateCallbackArgs = None

    def Print( s, text, timeout=5000, baton=False ):
        """ Print text to statusbar. """
        # This pends a change to statusbar text (rendering the widget is delayed).
        # The message is automatically cleared after a few seconds.
        # Pass timeout=0 to make the message stay indefinitely.
        # Pass baton=True inside a loop to twirl a baton (the loop drives the twirling).
        # -- don't call any GUI method, this can be run by second thread --
        raise Oops( "abstract method" )

    def PrintError( s, text ):
        """ Print error message. """
        raise Oops( "abstract method" )
        
    def PrintWarning( s, text ):
        """ Print warning message. """
        raise Oops( "abstract method" )
        
    def Clear( s ):
        """ Clear statusbar. """
        raise Oops( "abstract method" )

    def SetUpdateCallback( s, callback, *callbackArgs ):
        # This is to help periodically update the statusbar.
        # As the statusbar widget is updated by a timer,
        # SetUpdateFunc() can be called to specify a function
        # to be when the timer occurs and can call Statusbar.Print().
        # Pass None to set no function.
        st.gui.mutex.acquire()
        s.updateCallback     = callback
        s.updateCallbackArgs = callbackArgs
        st.gui.mutex.release()

#-------------------------------------------------------------------------------
# Class ScrollFrameBase.
#-------------------------------------------------------------------------------

class ScrollFrameBase(object):
    """ Create a frame and scrollbars to contain a widget. """

    def __init__( s, parent, WidgetClass, labelText, **args ):
        raise Oops( "abstract method" )

    def SetLabelText( s, text ):
        """ Set the text of a label widget in a scroll-frame widget. """
        raise Oops( "abstract method" )
    
    def SetLabelColor( s, color ):
        """ Set the color of a label widget in a scroll-frame widget. """
        raise Oops( "abstract method" )

    def GetScrollbarPos( s ):
        """ Returns (x,y) tuple of position of horizontal and vertical scrollbars. """
        raise Oops( "abstract method" )
    
    def SetScrollbarPos( s, xy ):
        """ Set position of horizontal and vertical scrollbars.  xy must be (x,y) tuple. """
        raise Oops( "abstract method" )
        
#-------------------------------------------------------------------------------
# Class NotebookBase.
#-------------------------------------------------------------------------------

class NotebookBase(object):
    """ Base class for a notebook widget. """
    # Page names are not the same as its tab text,
    # as pypersrc could browse files with the same names but different paths.

    def __init__( s, parentWidget ):
        pass

    def Enable( s, enable ):
        raise Oops( "abstract method" )

    def IfEmpty( s ):
        """ Return true if no pages. """
        raise Oops( "abstract method" )
    
    def AddPage( s, pageName, tabText, append=True ):
        """ Add a new notebook page.  Given page name must be unique.  tabText is for display.
            Returns widget so that caller can pack a child widget into it. """
        raise Oops( "abstract method" )

    def SelectPage( s, pageName ):
        """ Select a page. """
        raise Oops( "abstract method" )
        
    def RemovePage( s, pageName ):
        """ Remove a notebook page. """
        raise Oops( "abstract method" )

    def GetCurrentPageName( s ):
        """ Return name of current page or None if no pages exist. """
        raise Oops( "abstract method" )
        
    def CloseCurrentPage( s ):
        """ Close the current page. """
        raise Oops( "abstract method" )

#-------------------------------------------------------------------------------
# Class NotebookTextWid.
#-------------------------------------------------------------------------------

class NotebookTextWidBase(NotebookBase):

    def __init__( s ):
        pass

    def LoadContents( s, contents, line=0, hiliteLine=False, currentPage=False ):
        """ Load text widget (contained by notebook page) with Contents object. """
        # - line : Lines are counted from 1.  If line > 0, text widget is scrolled to line.
        raise Oops( "abstract method" )
    
    def GetCurrentTextWid( s ):
        """ Return text widget whose page is currently selected. """
        raise Oops( "abstract method" )

    def GetTextWidList( s ):
        """ Return list of text widgets in notebook. """
        raise Oops( "abstract method" )
    
#-------------------------------------------------------------------------------
# Class NotebookSrcFileBase.
#-------------------------------------------------------------------------------

class NotebookSrcFileBase(NotebookTextWidBase):

    def __init__( s ):
        pass
    
    def LoadSrcFile( s, fname, line=0 ):
        """ Load src file into notebook page. """
        raise Oops( "abstract method" )
        
    def Reload( s ):
        """ Reload src files in the notebook. """
        raise Oops( "abstract method" )
