# $Id: GuiText.py,v 1.51 2006-03-26 03:03:32 devel Exp $
# pypersrc
# Author:   Jim Brooks  http://www.jimbrooks.org
# LEGAL:    COPYRIGHT (C) JIM E. BROOKS 2005
#           THIS SOURCE CODE IS RELEASED UNDER THE TERMS
#           OF THE GNU GENERAL PUBLIC LICENSE (GPL).
# Notes:
# ==============================================================================

#-------------------------------------------------------------------------------
# These are base and fully-implemented classes, indepedent of any GUI toolkit,
# providing an interface to text widgets and their contents.
#
# fname
# -----
# A fname is associated with and identifies Contents objects and text widgets.
# fname can be a real filename or a phony one such as "[results: ]".
#
# Class Contents:
# ---------------
# The contents and state of a text widget is organized as a class
# so that it can be stored into ResultsHistory and CacheContents.
#
# Class TextWid:
# --------------
# This is derived from ScrollFrame and is very specialized for pypersrc.
# The top one is generally dedicated to holding source code, while the bottom one
# is for being loaded by hyperlinks, results from user commands, showing logs, etc.
#
# Class CacheSrcFile:
# -------------------
# A potentially slow operation is tokenizing a src file (tokens are stored
# as the Contents.tokenList member which is assigned by ContentsSrcFile constructor).
# Since the user can re-open and re-filter the same src files, caching is used
# to avoid retokenizing (file buffers, of course, are already cached by the OS).
# The cache item is keyed by both a fname and parse modes (modes being plural).
#
# Class ResultsHistory:
# ---------------------
# The results of commands are saved/cached in a history.
# Because refreshing can cause results to become stale, ResultsHistory
# has the ability to redo a command to freshen stale results.
# For example, if the user did a global search, then modified a src file,
# the global search command will be redone if the command's results
# are to be pulled from the history.
#
#-------------------------------------------------------------------------------
# Mechanisms:
#
# Caching Contents objects:
# -------------------------
# Contents objects contain a token list.  The process of tokeniziation can be slow.
# Text widgets display Contents objects.  Since Contents objects can migrate
# in and out of text widgets, caching Contents objects speeds reloading text widgets.
# ContentsSrcFile objects are cached by CacheSrcFile (see Class ContentsSrcFile).
# ContentsResults objects are cached by ResultsHistory (see Class ResultsHistory).
#
# Refreshing:
# -----------
# The user is supposed to click the refresh button after changing any source file
# (with an editor, removing, etc).  A goal is to avoid discarding or redoing results
# that are associated with unchanged src files.
#
# Periodic log refreshing:
# ------------------------
# Every second or so, IdleLog() is called.  If the log buffer has grown more than
# the log being shown in the text widget, the difference (the recently logged lines)
# is appended into the results text widget.
#-------------------------------------------------------------------------------

#-------------------------------------------------------------------------------
# Class Contents*.
#-------------------------------------------------------------------------------

class ContentsBase(object):
    """ The contents and state of a text widget. """

    def __init__( s, tokenList, fname ):
        # The contents stored is a list of tokens derived from the original source.
        s.tokenList = tokenList     # tokens from parsing source code or manufactured for results
        s.fname     = fname         # filename (can be phony)
        s.editable  = False         # whether contents is read-only or editable
        s.error     = False         # error during construction (invalid object)

    def IfEmpty( s ):
        """ Return true if no contents. """
        return not s.tokenList

    def IfError( s ):
        return s.error

class ContentsNone(ContentsBase):
    """ No contents. """
    # Why ContentsNone instead of using Python's None?
    # Client code assumes a TextWid (whether empty or not) has a Contents object.

    def __init__( s, tokenList=[], fname=None ):
        # fname can specified to mean empty file and to be shown in page tab.
        # tokenList can be specified too, but contents is still considered to have nothing important.
        ContentsBase.__init__( s, tokenList=tokenList, fname=fname )

class ContentsSrcFile(ContentsBase):
    """ Contents of a tokenized source file. """

    def __init__( s, fname ):
        ContentsBase.__init__( s, tokenList=[], fname=fname )
        s.fnameSaved = None    # fnameSaved/missing are for if a deleted src file-
        s.missing    = False   # -is reinstated (unusual case)

        # Read src file, tokenize it.
        st.log.Write( "ContentsSrcFile: reading/tokenizing src file " + fname )
        (s.tokenList,ok) = SrcFile2TokenList( fname )
        s.error = not ok
        s.parseModes = st.parser.GetModes()   # used by CacheSrcFile

class ContentsResults(ContentsBase):
    """ Contents from results dynamically computed by pypersrc.
        Appropriate ContentsResults can be updated by refreshing. """

    def __init__( s, tokenList, fname, command ):
        ContentsBase.__init__( s, tokenList, fname )
        s.command       = command    # user command that produced the results (N/A if this is a source file)
        s.refreshCnt    = st.refreshCnt    # to recompute stale results
        s.saveInHistory = True
        
    def UpdateRefreshCnt( s ):
        s.refreshCnt = st.refreshCnt

class ContentsSearch(ContentsResults):
    """ Results of a text search.  Might be affected by refreshing. """
    
    def __init__( s, tokenList, fname, command ):
        ContentsResults.__init__( s, tokenList, fname, command )

class ContentsLog(ContentsResults):
    """ For displaying log.  Automatically updated timer when log is being displayed. """
    
    def __init__( s, tokenList, fname ):
        ContentsResults.__init__( s, tokenList, fname, command="log" )

class ContentsHistory(ContentsResults):
    """ For displaying history of commands. """
    
    def __init__( s, tokenList, fname ):
        ContentsResults.__init__( s, tokenList, fname, command=None )

#-------------------------------------------------------------------------------
# Class TextWidBase.
#-------------------------------------------------------------------------------

class TextWidBase(ScrollFrameBase):
    """ Text widget in a scroll-frame widget. """
    
    def __init__( s ):
        # ContentsNone provides the required empty Contents.tokenList/.fname members so that
        # clearing a text widget can be done by passing ContentsNone to LoadContents().
        s.contents = ContentsNone()

    # INTERFACE
    def Enable( s, enable=True ):
        """ Enable text widget. """
        raise Oops( "abstract method" )

    # INTERFACE
    def IfEmpty( s ):
        """ Return True if text widget has no contents. """
        return isinstance( s.contents, ContentsNone )

    # INTERFACE
    def GetFname( s ):
        """ Return fname that identifies text widget. """
        return s.contents.fname

    # INTERFACE
    def LoadContents( s, contents, line=0, hiliteLine=False, scrollbarPos=None ):
        """ Load text widget with a Contents object. """
        # - contents : To implicitly force a reload, pass a different Contents object.
        # - scrollbarPos : (x,y) tuple (or None) of positions of horizontal,vertical scrollbars
        raise Oops( "abstract method" )
        
    # INTERFACE
    def GotoLine( s, line, hiliteLine=False ):
        raise Oops( "abstract method" )
        
    # INTERFACE
    def Refresh( s, change, addedList, modifiedList, removedList ):
        """ Refresh text widget after source files/dirs have changed. """
        raise Oops( "abstract method" )
        
#-------------------------------------------------------------------------------
# Class CacheSrcFile.
#-------------------------------------------------------------------------------

# Parse modes and filter modes are almost synonymous.
# Src files are filtered by setting the parse modes of Parser object.
    
class CacheSrcFile(object):
    """ Cache specialized for ContentsSrcFile objects. """
    # Src files can be filtered in different ways (parse modes).
    # Conceptually, ContentsSrcFile objects should be keyed by fname,parseModes.
    # How this is implemented in Python is that every fname is a dict key
    # which maps to a list of ContentsSrcFiles object which have the same fname
    # but have different parse modes.  This mapping scheme allows flushing cache
    # by fnames or parse modes.

    def __init__( s ):
        s.dictContentsList = {}    # fname : [ ContentsSrcFile ]

    # INTERFACE
    def Get( s, fname, parseModes ):
        """ Try to get ContentsSrcFile object keyed by fname,parseModes.
            If not cached, reads src file, caches and returns new ContentsSrcFile. """
        # Is it in the cache?
        (contentsFound,i) = s.__Get( fname, parseModes )
        if contentsFound is not None:
            # In the cache.
            st.log.Write( "CacheSrcFile: found " + s.__LogItem(fname,parseModes) + " in cache" )
            return contentsFound
        else:
            # Not in the cache.  Read src file from filesystem.
            st.log.Write( "CacheSrcFile: " + s.__LogItem(fname,parseModes) + " isn't cached" )
            contents = ContentsSrcFile( fname )
            if contents.IfError():
                raise IOError    # file read error
            else:
                # Store it in the cache.
                s.Put( contents, parseModes )
                return contents

    # PRIVATE
    def __Get( s, fname, parseModes ):
        """ Find ContentsSrcFile object keyed by fname,parseModes.
            Returns (contents,contentsListIndex) or (None,None). """
        if s.dictContentsList.has_key( fname ):
            # Ok, found fname in dict.
            # Now find parseModes in list (the ContentsSrcFile in desired filter mode).
            i = 0
            contentsList = s.dictContentsList[fname]
            for contents in contentsList:
                if parseModes == contents.parseModes:
                    return (contents,i)
                i += 1
        return (None,None)

    # INTERFACE
    def Put( s, contents, parseModes ):
        """ Put a ContentsSrcFile object into cache. """
        # This cache is only for ContentsSrcFiles objects.
        if isinstance(contents,ContentsSrcFile):
            st.log.Write( "CacheSrcFile: caching " + s.__LogItem(contents.fname,parseModes) )
            s.__Put( contents, parseModes )
        elif isinstance(contents,ContentsNone):
            return    # silently ignore, normally happens
        else:
            raise Oops( "CacheSrcFile.Put: wrong Contents object to cache" )

    # PRIVATE
    def __Put( s, contents, parseModes ):
        """ Put a ContentsSrcFile into cache (directly, no checks). """
        fname = contents.fname
        (contentsFound,i) = s.__Get( fname, parseModes )
        if contentsFound:
            # Found.
            # Does cached contents differ from passed contents?
            if contents is contentsFound:
                return                                      # same contents, was already in cache
            else:
                contentsList = s.dictContentsList[fname]    # different contents, update cache
                contentsList[i] = contents                  # replace old contents
                return
        else:
            # Mismatch.
            if s.dictContentsList.has_key( fname ):
                # Partial match:  fname is in the dict but has no matching parse modes.
                # Append this contents to list.
                contentsList = s.dictContentsList[fname] 
                contentsList.append( contents )
                return
            else:
                # Total mismatch.  This fname was never encountered before.
                # Create new contents list for fname.
                s.dictContentsList[fname] = [ contents ]
                return

    # INTERFACE
    def Flush( s, fname=None, parseModes=None ):
        """ Flush by either fname or parseModes.
            Flush entire cache if neither fname nor parseModes. """
        # If cache is empty, nothing to flush.
        if s.dictContentsList:
            if (fname is None) and (parseModes is None):    # "not parseModes" is wrong 
                # Flush everything.                         # as parseModes=0 is possible
                st.log.Write( "CacheSrcFile: flushing entire cache" )
                #del s.dictContentsList    # wrong FYI
                s.dictContentsList = {}
            elif (fname is not None) and (parseModes is None):
                # Flush all ContentsSrcFile objects matching fname (all of its parse modes).
                if s.dictContentsList.has_key( fname ):
                    st.log.Write( "CacheSrcFile: flushing " + fname + " (all modes)" )
                    del s.dictContentsList[fname]
            elif (fname is None) and (parseModes is not None):
                # Flush all ContentsSrcFile objects matching parse modes (any fname).
                flushed = False
                for contentsList in s.dictContentsList.values():
                    i = 0
                    for contents in contentsList:
                        if contents.parseModes == parseModes:
                            del contentsList[i]   # ok as break follows deletion
                            flushed = True
                            break
                        i += 1
                if flushed:
                    st.log.Write( "CacheSrcFile: flushed src files in " + st.parser.GetModesString(parseModes) )
            else:
                raise Oops( "CacheSrcFile.Flush: unimplemented case" )    # no need for this case

    # PRIVATE
    def __LogItem( s, fname, parseModes ):
        """ Helper to abbrev logging.  Returns string. """
        return fname + " " + st.parser.GetModesString(parseModes)

#-------------------------------------------------------------------------------
# Class ResultsHistory.
#-------------------------------------------------------------------------------

# textWid args should be a TextWidResults object (not the base TextWid).

class ResultsHistory(object):
    """ Maintains a history of results. """

    def __init__( s ):
        s.contentsList        = []     # maintains order
        s.contentsDict        = {}     # fname : Contents
        s.scrollbarPosDict    = {}     # fname : scrollbar position (x,y) tuple
        s.resultsHistoryFname = "[results: history of command results]"

    # INTERFACE
    def Save( s, textWid ):
        """ Store the results (being displayed in text widget) into history. """
        # Handles duplicates (because of how Show() works).
        # The command history itself shouldn't be saved.
        # Yes, ContentsLog should be saved in history.
        contents = textWid.contents
        if isinstance(contents,ContentsResults) and not isinstance(contents,ContentsHistory) and contents.saveInHistory:
            # Save contents and scrollbar position.
            # Don't save duplicates.
            fname = contents.fname
            if not s.contentsDict.has_key( fname ):
                s.contentsList.append( contents )
                s.contentsDict[fname] = contents
            scrollbarPos = textWid.GetScrollbarPos()  # returns (x,y) tuple
            s.scrollbarPosDict[fname] = scrollbarPos

    # INTERFACE
    def Load( s, textWid, fname ):
        """ Load text widget with results from history. """
        # Has results become stale because of recent refreshing?
        contents = s.contentsDict[fname]
        if contents.refreshCnt < st.refreshCnt and contents.command:
            # Redo command to refresh results.
            contents.refreshCnt = st.refreshCnt
            st.log.Write( "ResultsHistory: stale results, redoing command '" + contents.command + "'" )
            st.command.Do( contents.command )
            # Command.Do() has again loaded the results text widget.
        else:
            # Load text widget with saved results, restore scrollbar position.
            scrollbarPos = s.scrollbarPosDict[fname]
            textWid.LoadContents( contents=contents, scrollbarPos=scrollbarPos )

    # INTERFACE
    def Show( s, textWid ):
        """ Show the history of command results. """
        # The results history is shown as hyperlinked fnames to saved results.
        # The most recent results is shown at the top.

        # Save the current results right now to show it also.
        s.Save( textWid )

        # Build fname list.
        fnameList = []
        for contents in s.contentsList:
            fnameList.insert( 0, contents.fname )

        # Build token list from fname list.
        tokenList = []
        tokenList.append( (eTokenText,"History of command results:\n\n") )
        for fname in fnameList:
            phonyTag = st.tags.PhonyTag( fname, fname, lineNum=0 )    # hyperlink
            tokenList.append( (eTokenTag,phonyTag) )

        # Build contents from token list, then load it into the text widget.
        # Indeed, ResultsHistory.Save() will be called twice as TextWidResults.LoadContents()
        # also calls it.  But Save() handles being passed duplicate results.
        contents = ContentsHistory( tokenList=tokenList, fname=s.resultsHistoryFname )
        textWid.LoadContents( contents=contents )

    # INTERFACE
    def ShowLast( s, textWid ):
        """ Show the results from the last command. """
        # textWid is the text widget to show results in.
        if s.contentsList:
            fname = s.contentsList[-1].fname
            s.Load( textWid, fname )

    # INTERFACE
    def Refresh( s, change, addedList, modifiedList, removedList ):
        """ Refresh history after source files have changed. """
        # This has become a NOP as ResultsHistory.Load() refreshes stale results
        # (this is kept for logical completeness in HandlerButtonRefresh()).
        pass
