# $Id: Lex.py,v 1.36 2006-03-26 03:03:32 devel Exp $
# pypersrc
# Author:   Jim Brooks  http://www.jimbrooks.org
# LEGAL:    COPYRIGHT (C) JIM E. BROOKS 2005
#           THIS SOURCE CODE IS RELEASED UNDER THE TERMS
#           OF THE GNU GENERAL PUBLIC LICENSE (GPL).
# ==============================================================================

#-------------------------------------------------------------------------------
# Parsing source code for function calls, variable declarations, etc is simplified
# by first scanning thru the raw source code to produce "tokens" (this is the
# idea behind lex and yacc).  A parser can parse thru tokens rather than raw text.
# In this Python implementation, a token is a tuple consisting of an enumeration
# and the text that yielded the token.
#
# Class is-a:
#
# LexBase --> LexC
#             LexPython
#             LexPerl
#             
# Class has-a:
#
# Lex -->     LexC
#             LexPython
#             LexPerl
#-------------------------------------------------------------------------------

# Enums are prefixed by "e".
# 0x00..0xff  : chars with no corresponding tokens
# 0x100..0x1ff: tokens that can be combined if adjacent
# 0x200..0x2ff: tokens that can't be combined
# combinable:
eTokenInvalid     = 0x100
eTokenWs          = 0x101   # space, tab, newline
eTokenSymbol      = 0x102   # symbol/punctuation chars
eTokenNum         = 0x103   # numeric chars
eTokenWord        = 0x104   # a word or a possible identifier
eTokenKeyword     = 0x105   # a word or a possible identifier
eTokenKeywordLoud = 0x106   # a word or a possible identifier
eTokenType        = 0x107   # data type (int,bool)
eTokenText        = 0x108   # for results from commands (not used for src files)
# non-combinable:
eTokenTag         = 0x200   # a source code tag
eTokenQuoted      = 0x201   # text inside quotes  
eTokenComment     = 0x202   # comment (language specific)

# ------------------------------------------------------------------------------
# Class Lex.
# ------------------------------------------------------------------------------

class Lex(object):

    def __init__( s, argList ):
        s.ProcessArgsPre( argList )
        s.lexC        = LexC()        # includes C++
        s.lexJava     = LexJava()
        s.lexPython   = LexPython()
        s.lexLua      = LexLua()
        s.ProcessArgsPost( argList )
        return

    # INTERFACE
    def ProcessArgsPre( s, argList ):
        """ Assigns None to elements in argList that were processed. """
        i = 0
        while i < len(argList):
            arg = argList[i]
            if arg is None:
                i += 1
            elif arg.find( "-hilite-no" ) != -1:    # no/none
                LexBase.SetHilite( "no" )           # static method
                i += 1
            elif arg.find( "-hilite-lo" ) != -1:    # hi/high
                LexBase.SetHilite( "lo" )           # static method
                i += 1
            elif arg.find( "-hilite-hi" ) != -1:    # lo/low
                LexBase.SetHilite( "hi" )           # static method
                i += 1
            else:
                i += 1
        return True

    # INTERFACE
    def ProcessArgsPost( s, argList ):
        """ Assigns None to elements in argList that were processed. """
        return True

    # INTERFACE
    def LexFile( s, fname, buf="" ):
        """ Lex a source code file.  Returns a list of token tuples. """

        # If buffer wasn't passed, open file.
        if buf == "":
            try:
                buf = ReadSrcFile( fname )
            except IOError:
                PrintWarning( "Lex: can't open file " + fname )
                return []

        proglang = WhichKindOfSrcFile( fname )

        if proglang == "c" or proglang == "c++":
            return s.lexC.LexBuffer( buf )
        elif proglang == "java":
            return s.lexJava.LexBuffer( buf )
        elif proglang == "python":
            return s.lexPython.LexBuffer( buf )
        elif proglang == "lua":
            return s.lexLua.LexBuffer( buf )
        else:
            PrintWarning( "Lex: unknown source file" )
            return []

# ------------------------------------------------------------------------------
# Class LexBase.
# ------------------------------------------------------------------------------

class LexBase(object):
    """ Tokenizes source code for parsing. """

    # 0=none 1=low 2=high
    hilite = 2

    # Precompute token table that correlates ASCII chars to eToken enums.
    # The token table is language-specific.
    tokenTable1 = []    # more hiliting (slower)
    tokenTable2 = []    # less hiliting (faster)
    for i in range(0,0x100):
        tokenTable1.append(0)
    for i in range(33,48):
        tokenTable1[i] = eTokenText
    for i in range(48,58):
        tokenTable1[i] = eTokenWord #eTokenNum  # treat numbers as alphabetic 
    for i in range(58,65):                       # to handle alpha-numeric tags
        tokenTable1[i] = eTokenSymbol
    for i in range(65,91):
        tokenTable1[i] = eTokenWord
    for i in range(91,97):
        tokenTable1[i] = eTokenSymbol
    for i in range(97,123):
        tokenTable1[i] = eTokenWord
    for i in range(123,127):
        tokenTable1[i] = eTokenSymbol
    for c in " \t\n\r":
        tokenTable1[ord(c)] = eTokenText
    tokenTable1[ord("_")] = eTokenWord
    # tokenTable2 is for low hiliting (lacks some tokens).
    tokenTable2 = tokenTable1[:]
    for i in range(0,0x100):
        if tokenTable2[i] in [eTokenSymbol,eTokenWs]:
            tokenTable2[i] = eTokenText
    tokenTable = tokenTable2

    def __init__( s ):

        # Here's Pythonic programming:
        # A base constructor is calling a method in a derived class (impossible in C++).
        # This is cleaner than rewriting it the C++ way by passing down keywordDict arg.
        keywordDict = s.KeywordDict()

        # Holds contents of a source file.
        s.deq = DequeChar()

        # Prepare for lexing keywords.
        s.keywordMin = 1000
        s.keywordMax = 0
        for keyword in keywordDict.keys():
            keywordLen = len(keyword)
            if keywordLen < s.keywordMin:
                s.keywordMin = keywordLen
            elif keywordLen > s.keywordMax:
                s.keywordMax = keywordLen

        # Used to match keywords using "greedy" searches.
        # Eg, to find "#ifdef" instead of just "#if".
        s.greedyDictKeywords = GreedyDict( s.keywordMax )
        for (key,val) in keywordDict.items():
            s.greedyDictKeywords[key] = val
        
        return

    # INTERFACE
    def LexBuffer( s, buf ):
        """ Lex a buffer (string).  Returns a list of token tuples. """

        # Use C++ STL deque which is more suitable and faster than Python strings.
        s.deq.Set( buf )
        
        # Ensure file's last char is EOL which eliminates special cases.
        if s.deq.PeekTail() != "\n":
            s.deq.PushTail( "\n" )
            
        tokenList = []
        token     = eTokenInvalid
        tokenPrev = eTokenInvalid
        lineNum   = 0

        # No hilite is simple.
        if LexBase.hilite == 0:
            try:
                text = s.deq.Get()
                t = (token,text)
                tokenList = [ t ]
            except:
                PrintException()
            return tokenList
        
        # First pass.
        try:
            while not s.deq.IfEmpty():
                c = s.deq.PeekHead()
                # Keyword?
                text = ""
                if LexBase.hilite == 2:
                    if tokenPrev != eTokenWord:  # avoid misidentifying a word containing a keyword
                        (token,text) = s.LexKeyword( tokenPrev )
                if text:
                    pass    # (token,text) already
                # Quoted string?
                elif c == '"' or c == "'":
                    token = eTokenQuoted
                    text  = s.LexQuotedString()
                # Comment?
                elif s.IfCommentBegin():
                    # Comments are lang-specific, defer to derived methods.
                    token = eTokenComment
                    text  = s.LexComment()
                # EOL?
                elif c == "\n":
                    s.deq.PopHead()
                    lineNum += 1
                    token = eTokenText
                    text  = c
                # Token in table?
                elif LexBase.tokenTable[ord(c)]:
                    s.deq.PopHead()
                    token = LexBase.tokenTable[ord(c)]
                    text  = c
                # Uncategorized token.
                else:
                    s.deq.PopHead()
                    continue    # ignore it

                # Combine the same kind of tokens (not all kinds can be combined).
                if token == tokenPrev and s.IfCombinableToken(token):
                    tokenList[-1] = (token,tokenList[-1][1] + text)
                    continue

                # Append tuple to token list.
                t = (token,text)
                tokenList.append( t )
                tokenPrev = token
        except:
            PrintException()

        # Second pass: Find words that really are tags.
        if LexBase.hilite:
            try:
                tokenList2 = []
                while tokenList:
                    token = tokenList[0][0]
                    text  = tokenList[0][1]
                    if token == eTokenWord and pycpp_tags.IfTag( text ):
                        token = eTokenTag    # really is a tag
                    tokenList2.append( (token,text) )
                    del tokenList[0]
            except:
                PrintException()
            tokenList = tokenList2

        # Third pass: Correct problem of false keywords.
        if LexBase.hilite == 2:
            try:
                tokenList = s.CorrectFalseKeywords( tokenList )
            except:
                PrintException()

        # Fourth pass: Combine tokens (again).
        if LexBase.hilite == 1:
            try:
                prev = (-999,"")
                tokenList2 = []
                i = 0
                j = 0
                for t in tokenList:
                    if t[0] == eTokenWord or t[0] == eTokenKeyword:
                        t = (eTokenText,t[1])

                    if t[0] != prev[0]:
                        tokenList2.append( t )
                        j += 1
                    else:
                        tokenList2[j-1] = (t[0], tokenList2[j-1][1] + t[1])
                    prev = t
                    i += 1
                tokenList = tokenList2
            except:
                PrintException()

        # Done.
        return tokenList

    # INTERFACE/STATIC
    def SetHilite( level ):  # self omitted
        # This is a Python static method -- eg LexBase.SetHilite("lo").
        if level in ["no","none"]:
            LexBase.tokenTable = LexBase.tokenTable2
            LexBase.hilite = 0
        elif level in ["lo","low"]:
            LexBase.tokenTable = LexBase.tokenTable2
            LexBase.hilite = 1
        elif level in ["hi","high"]:
            LexBase.tokenTable = LexBase.tokenTable1
            LexBase.hilite = 2
    SetHilite = staticmethod(SetHilite)

    # PROTECTED
    def IfCombinableToken( s, token ):
        return 0x100 <= token <= 0x1ff

    # PROTECTED
    def CorrectFalseKeywords( s, tokenList ):
        """ Eliminate false keywords.  Returns new tokenList. """
        # This is necessary because text such as "g_return_if_fail"
        # would be lexed incorrectly because "return" is a C keyword.
        # This method corrects such errors.
        # TODO For FORTH this would be different (in derived class).
        try:
            tokenList2 = []
            keywordList     = [eTokenKeyword,eTokenType]
            keywordLikeList = [eTokenWord,eTokenNum]   # tokens similar to keyword tokens
            while tokenList:
                token = tokenList[0][0]
                if not token in keywordLikeList and not token in keywordList:
                    tokenList2.append( tokenList[0] )
                    del tokenList[0]
                    continue                    
                # keywordlike, keyword, keywordlike
                if  tokenList[0][0] in keywordLikeList \
                and tokenList[1][0] in keywordList \
                and tokenList[2][0] in keywordLikeList:
                    token = tokenList[0][0]    # first is preferred, third may be number when first is alpha
                    text = tokenList[0][1] + tokenList[1][1] + tokenList[2][1]
                    tokenList2.append( (token,text) )
                    del tokenList[0:3]
                    continue
                # keywordlike, keyword
                if  tokenList[0][0] in keywordLikeList \
                and tokenList[1][0] in keywordList:
                    token = tokenList[0][0]
                    text = tokenList[0][1] + tokenList[1][1]
                    tokenList2.append( (token,text) )
                    del tokenList[0:2]
                    continue
                # keyword, keywordlike
                if  tokenList[0][0] in keywordList \
                and tokenList[1][0] in keywordLikeList:
                    token = tokenList[1][0]
                    text = tokenList[0][1] + tokenList[1][1]
                    tokenList2.append( (token,text) )
                    del tokenList[0:2]
                    continue
                # no mistaken keywords
                tokenList2.append( tokenList[0] )
                del tokenList[0]
        except:
            PrintException()
        return tokenList2
        
    # PROTECTED
    def LexQuotedString( s ):
        """ Lex a quoted string in source code. """

        quote = s.deq.PopHead()
        text = quote

        # Find the ending quote char.
        while not s.deq.IfEmpty():
            c = s.deq.PopHead()
            text += c
            # End quote?
            if c == quote:
                return text
            # Escaped char?
            elif c == "\\":
                # TODO handle more complex escaped cases.
                if not s.deq.IfEmpty():
                    text += s.deq.PopHead()
                
        # String wasn't terminated.
        return text

    # PROTECTED
    def LexKeyword( s, tokenPrev ):
        """ Lex a possible keyword.  Returns (token,keyword) or (None,None). """
        # Greedy match.
        try:
            try:
                text = s.deq.PeekHead( s.keywordMax )
            except:
                text = s.deq.PeekHead( s.deq.Size() )  # peeked past EOF
            (key,val) = s.greedyDictKeywords.Find( text )
            s.deq.PopHead( len(key) )
            return (val,key)
        except:
            return (None,None)

    # PROTECTED
    def LexDelim( s, delim ):
        """ Lex text til an end delimiter.
            Returns string including the end delimiter. """
        idx = s.deq.Index( delim )  # may raise exception 
        return s.deq.PopHead( idx + len(delim) )  # from beginning til end delimiter

# ------------------------------------------------------------------------------
# Class LexC (C/C++).
# ------------------------------------------------------------------------------

class LexC(LexBase):
    """ Lex class for C and C++. """

    def __init__( s ):
        LexBase.__init__( s )
        return

    # PROTECTED
    def KeywordDict( s ):
        return \
        { "return"            : eTokenKeyword,
          "if"                : eTokenKeyword,
          "else"              : eTokenKeyword,
          "return"            : eTokenKeyword,
          "for"               : eTokenKeyword,
          "while"             : eTokenKeyword,
          "do"                : eTokenKeyword,
          "continue"          : eTokenKeyword,
          "break"             : eTokenKeyword,
          "switch"            : eTokenKeyword,
          "case"              : eTokenKeyword,
          "default"           : eTokenKeyword,
          "sizeof"            : eTokenKeyword,
          "goto"              : eTokenKeywordLoud,
          "NULL"              : eTokenKeywordLoud,
          "asm"               : eTokenKeyword,
          # C pre-processsor
          "#define"           : eTokenKeyword,
          "#ifdef"            : eTokenKeyword,
          "#ifndef"           : eTokenKeyword,
          "#if"               : eTokenKeyword,
          "#else"             : eTokenKeyword,
          "#elif"             : eTokenKeyword,
          "#endif"            : eTokenKeyword,
          #  
          "#\tdefine"         : eTokenKeyword,    # not ideal but would be
          "#\tifdef"          : eTokenKeyword,    # slow/difficult as a special-case
          "#\tifndef"         : eTokenKeyword,
          "#\tif"             : eTokenKeyword,
          "#\telse"           : eTokenKeyword,
          "#\telif"           : eTokenKeyword,
          "#\tendif"          : eTokenKeyword,
          #  
          "#\t\tdefine"       : eTokenKeyword,
          "#\t\tifdef"        : eTokenKeyword,
          "#\t\tifndef"       : eTokenKeyword,
          "#\t\tif"           : eTokenKeyword,
          "#\t\telse"         : eTokenKeyword,
          "#\t\telif"         : eTokenKeyword,
          "#\t\tendif"        : eTokenKeyword,
          #  
          "#   define"        : eTokenKeyword,
          "#   ifdef"         : eTokenKeyword,
          "#   ifndef"        : eTokenKeyword,
          "#   if"            : eTokenKeyword,
          "#   else"          : eTokenKeyword,
          "#   elif"          : eTokenKeyword,
          "#   endif"         : eTokenKeyword,
          #  
          "#       define"    : eTokenKeyword,    # ok, that's enough
          "#       ifdef"     : eTokenKeyword,
          "#       ifndef"    : eTokenKeyword,
          "#       if"        : eTokenKeyword,
          "#       else"      : eTokenKeyword,
          "#       elif"      : eTokenKeyword,
          "#       endif"     : eTokenKeyword,
          #
          "#include"          : eTokenKeyword,
          "#undef"            : eTokenKeyword,
          "#error"            : eTokenKeyword,
          "#pragma"           : eTokenKeyword,
          "defined"           : eTokenKeyword,
          # C++
          "catch"             : eTokenKeyword,
          "delete"            : eTokenKeyword,
          "using"             : eTokenKeyword,
          "namespace"         : eTokenKeyword,
          "new"               : eTokenKeyword,
          "operator"          : eTokenKeyword,
          "private"           : eTokenKeyword,
          "public"            : eTokenKeyword,
          "this"              : eTokenKeyword,
          "throw"             : eTokenKeyword,
          "try"               : eTokenKeyword,
          "typeid"            : eTokenKeyword,
          "typename"          : eTokenKeyword,
          "static_cast"       : eTokenKeywordLoud,
          "const_cast"        : eTokenKeywordLoud,
          "reinterpret_cast"  : eTokenKeywordLoud,
          # C types
          "void"              : eTokenType,
          "char"              : eTokenType,
          "short"             : eTokenType,
          "int"               : eTokenType,
          "long"              : eTokenType,
          "float"             : eTokenType,
          "double"            : eTokenType,
          "const"             : eTokenType,
          "volatile"          : eTokenType,
          "register"          : eTokenType,
          "static"            : eTokenType,
          "extern"            : eTokenType,
          "signed"            : eTokenType,
          "unsigned"          : eTokenType,
          "struct"            : eTokenType,
          "union"             : eTokenType,
          "typedef"           : eTokenType,
          "enum"              : eTokenType,
          # C++ types
          "bool"              : eTokenType,
          "class"             : eTokenType,
          "explicit"          : eTokenType,
          "export"            : eTokenType,
          "friend"            : eTokenType,
          "mutable"           : eTokenType,
          "protected"         : eTokenType,
          "template"          : eTokenType,
          "virtual"           : eTokenType,
          "inline"            : eTokenType,
          # C++ STL types
          "string"            : eTokenType,
          "vector"            : eTokenType,
          "deque"             : eTokenType,
          "set"               : eTokenType,
          "multiset"          : eTokenType,
          "map"               : eTokenType,
          "multimap"          : eTokenType,
          "iterator"          : eTokenType,
          "const_iterator"    : eTokenType,
          "size_type"         : eTokenType,
        }
    
    # PROTECTED
    def IfCommentBegin( s ):
        """ Return True if this is the start of a comment. """
        try:
            text = s.deq.PeekHead( 2 )
            if text[0] != "/":
                return False
            if text[1] == "*":  # C comment?
                return True
            if text[1] == "/":  # C++ comment?
                return True
            return False
        except:
            return False

    # PROTECTED
    def LexComment( s ):
        """ Lex a comment.
            @pre deq is pointing to the start of a comment. """
        text = s.deq.PeekHead( 2 )
        if text == "/*":
            return s.LexDelim( "*/" )
        elif text == "//":
            return s.LexDelim( "\n" )
        else:
            raise Oops

# ------------------------------------------------------------------------------
# Class LexJava.
# ------------------------------------------------------------------------------

class LexJava(LexC):

    def __init__( s ):
        LexC.__init__( s )
        return

    # PROTECTED
    def KeywordDict( s ):
        return \
        { "abstract"          : eTokenKeyword,   
          "break"             : eTokenKeyword,
          "case"              : eTokenKeyword,
          "catch"             : eTokenKeyword,
          "class"             : eTokenKeyword,
          "continue"          : eTokenKeyword,
          "default"           : eTokenKeyword,
          "do"                : eTokenKeyword,
          "else"              : eTokenKeyword,
          "extends"           : eTokenKeyword,
          "false"             : eTokenKeyword,
          "finally"           : eTokenKeywordLoud,
          "for"               : eTokenKeyword,
          "if"                : eTokenKeyword,
          "implements"        : eTokenKeyword,
          "import"            : eTokenKeyword,
          "instanceof"        : eTokenKeyword,
          "interface"         : eTokenKeyword,
          "native"            : eTokenKeyword,
          "new"               : eTokenKeyword,
          "null"              : eTokenKeywordLoud,
          "package"           : eTokenKeyword,
          "return"            : eTokenKeyword,
          "strictfp"          : eTokenKeyword,
          "super"             : eTokenKeyword,
          "switch"            : eTokenKeyword,
          "synchronized"      : eTokenKeyword,
          "this"              : eTokenKeyword,
          "throw"             : eTokenKeywordLoud,
          "throws"            : eTokenKeyword,
          "transient"         : eTokenKeyword,
          "true"              : eTokenKeyword,
          "try"               : eTokenKeyword,
          "while"             : eTokenKeyword,
          # types
          "private"           : eTokenType,
          "protected"         : eTokenType,
          "public"            : eTokenType,
          "static"            : eTokenType,
          "final"             : eTokenType,
          "boolean"           : eTokenType,
          "byte"              : eTokenType,
          "char"              : eTokenType,
          "short"             : eTokenType,
          "int"               : eTokenType,
          "long"              : eTokenType,
          "float"             : eTokenType,
          "double"            : eTokenType,
          "const"             : eTokenType,
          "void"              : eTokenType,
          "volatile"          : eTokenType,
          "String"            : eTokenType,
        }

# ------------------------------------------------------------------------------
# Class LexScript.
# ------------------------------------------------------------------------------

class LexScript(LexBase):
    """ Base class for languages that use # for comments. """
    
    def __init__( s ):
        LexBase.__init__( s )
        return

    # PROTECTED
    def IfCommentBegin( s ):
        """ Return True if this is the start of a comment. """
        try:
            text = s.deq.PeekHead( 1 )
            return text[0] == "#"
        except:
            return False

    # PROTECTED
    def LexComment( s ):
        """ Lex a comment.
            @pre deq is pointing to the start of a comment. """
        return s.LexDelim( "\n" )

# ------------------------------------------------------------------------------
# Class LexPython.
# ------------------------------------------------------------------------------

class LexPython(LexScript):

    def __init__( s ):
        LexBase.__init__( s )
        return

    # PROTECTED
    def KeywordDict( s ):
        return \
        { "return"          : eTokenKeyword,
          "and"             : eTokenKeyword,
          "assert"          : eTokenKeyword,
          "break"           : eTokenKeyword,
          "class"           : eTokenKeyword,
          "continue"        : eTokenKeyword,
          "def"             : eTokenKeyword,
          "del"             : eTokenKeyword,
          "elif"            : eTokenKeyword,
          "else"            : eTokenKeyword,
          "except"          : eTokenKeyword,
          "exec"            : eTokenKeyword,
          "finally"         : eTokenKeyword,
          "for"             : eTokenKeyword,
          "from"            : eTokenKeyword,
          "global"          : eTokenKeyword,
          "if"              : eTokenKeyword,
          "import"          : eTokenKeyword,
          "in"              : eTokenKeyword,
          "is"              : eTokenKeyword,
          "lambda"          : eTokenKeyword,
          "not"             : eTokenKeyword,
          "or"              : eTokenKeyword,
          "pass"            : eTokenKeyword,
          "print"           : eTokenKeyword,
          "raise"           : eTokenKeywordLoud,
          "return"          : eTokenKeyword,
          "try"             : eTokenKeyword,
          "while"           : eTokenKeyword,
        }

# ------------------------------------------------------------------------------
# Class LexLua.
# ------------------------------------------------------------------------------

class LexLua(LexBase):

    def __init__( s ):
        LexBase.__init__( s )
        return

    # PROTECTED
    def KeywordDict( s ):
        return \
        { "and"             : eTokenKeyword,
          "break"           : eTokenKeyword,
          "do"              : eTokenKeyword,
          "else"            : eTokenKeyword,
          "elseif"          : eTokenKeyword,
          "end"             : eTokenKeyword,
          "error"           : eTokenKeywordLoud,
          "for"             : eTokenKeyword,
          "function"        : eTokenKeyword,
          "if"              : eTokenKeyword,
          "local"           : eTokenKeyword,
          "nil"             : eTokenKeywordLoud,
          "not"             : eTokenKeyword,
          "or"              : eTokenKeyword,
          "repeat"          : eTokenKeyword,
          "return"          : eTokenKeyword,
          "then"            : eTokenKeyword,
          "until"           : eTokenKeyword,
          "while"           : eTokenKeyword,
        }

    # PROTECTED
    def IfCommentBegin( s ):
        """ Return True if this is the start of a comment. """
        try:
            text = s.deq.PeekHead( 2 )
            return text[0] == "-" and text[1] == "-"
        except:
            return False

    # PROTECTED
    def LexComment( s ):
        """ Lex a comment.
            @pre deq is pointing to the start of a comment. """
        return s.LexDelim( "\n" )

