# $Id: Parse.py,v 1.15 2006-03-26 03:03:32 devel Exp $
# pypersrc
# Author:   Jim Brooks  http://www.jimbrooks.org
# LEGAL:    COPYRIGHT (C) JIM E. BROOKS 2005
#           THIS SOURCE CODE IS RELEASED UNDER THE TERMS
#           OF THE GNU GENERAL PUBLIC LICENSE (GPL).
# ==============================================================================

#-------------------------------------------------------------------------------
# Class Parse.
#-------------------------------------------------------------------------------

# Parse modes are set by bitwise combinations.  The parser can be in multiple modes.
# Src files are filtered by setting the parse modes of Parser object.
eParseDefault   = 0
eParseComments  = 1
eParseFunctions = 2
eParseData      = 4
parseModesNameDict = { eParseComments  : "comment",
                       eParseFunctions : "function",
                       eParseData      : "data" }
# Bitwise combination of parse modes that filter source code.
filterParseModes = eParseComments | eParseFunctions | eParseData

class Parse(object):
    """ Parse tokenized source code. """
    # See also Lex.py.

    def __init__( s, argList ):
        s.ProcessArgsPre( argList )
        s.modes = 0                          # bitwise combination
        s.ProcessArgsPost( argList )
        return

    # INTERFACE
    def ProcessArgsPre( s, argList ):
        """ Assigns None to elements in argList that were processed. """
        return True
    
    # INTERFACE
    def ProcessArgsPost( s, argList ):
        """ Assigns None to elements in argList that were processed. """
        return True

    # INTERFACE
    def IfFiltering( s ):
        """ Return True if the parser is filtering source code elements. """
        return s.modes & filterParseModes

    # INTERFACE
    def GetModesString( s, parseModes=None ):
        """ Return modes as a printable/informational string. """
        # Eg builds "functions,comments" or "default".
        if parseModes is None:
            parseModes = s.modes
        modesStr = ""
        for modeBit in parseModesNameDict.keys():
            if modeBit & parseModes:
                modesStr += parseModesNameDict[modeBit] + ","
        if modesStr:
            modesStr = modesStr[0:-1]    # chop final ","
        else:
            modesStr = "default"
        return "(modes=" + modesStr + ")"

    # INTERFACE
    def GetModes( s ):
        """ Return modes (bitwise int) the parser is current in. """
        return s.modes

    # INTERFACE
    def SetModes( s, modes ):
        """ Set the modes the parser will be in.  modes is a list of eParse* enums. """
        # modes should be bitwise int combination.
        s.modes = s.modes

    # INTERFACE
    def SetDefaultMode( s ):
        """ Set the parser default mode. """
        s.modes = 0

    # INTERFACE
    def AddMode( s, mode ):
        """ Put the parser into an additional mode. """
        s.modes = s.modes | mode

    # INTERFACE
    def RemoveMode( s, mode ):
        """ Take the parser out of a mode. """
        s.modes = s.modes & ~mode

    # INTERFACE
    def ParseFile( s, fname, tokenList ):
        """ Returns a second token list from the passed one. """
        # For example, to digest comments, the second token list would omit
        # most of the original tokens.

        # Show source file verbatim (not digested)?
        if not s.modes:
            return tokenList

        # Digest comments, functions, etc into a dictionary of lineNums/tokens.
        # Then sort and merge the digestions.
        # The line numbers are scaled, because, for example, a hyperlink tag
        # is a token as is the text token for its newline, but both are on the same line,
        # so the line number is scaled and the newline is given a line number one higher.
        digestedTokenDict = { }

        if eParseComments & s.modes:
            s.__DigestComments( fname, tokenList, digestedTokenDict )
        
        if    (eParseFunctions & s.modes) \
           or (eParseData      & s.modes):
            s.__DigestFunctionsOrData( fname, tokenList, digestedTokenDict )

        mergedTokenList = []
        keyList = digestedTokenDict.keys()
        keyList.sort()
        for key in keyList:
            mergedTokenList.append( digestedTokenDict[key] )
        return mergedTokenList

    # PRIVATE
    def __DigestComments( s, fname, tokenList, digestedTokenDict ):
        """ Adds comments to a dictionary of lineNums/tokens. """
        lineNum = 1
        for token in tokenList:
            if token[0] == eTokenComment:
                # See ParseFile() why line numbers are scaled.
                digestedTokenDict[lineNum*10+9] = (token[0], Chomp(token[1])+"\n") # token
            lineNum += token[1].count( "\n" )

    # PRIVATE
    def __DigestFunctionsOrData( s, fname, tokenList, digestedTokenDict ):
        """ Adds tags of functions or data to a dictionary of lineNums/tokens. """
        try:
            # Show a terse representation: show just a function on every line.
            # Get list of tags in this module.
            tagList = st.tags.GetModuleTags( fname )
            if tagList:
                # Build a dictionary of tags keyed by line number
                # in order to sort the tags by line number.
                tagDict = { }
                for tag in tagList:
                    tagDict[ tag[TAG_LINE] ] = tag

                # Build token list using tags sorted by line number.
                keyList = tagDict.keys()
                keyList.sort()
                for key in keyList:
                    tag = tagDict[key]
                    lineNumScaled = tag[TAG_LINE]*10
                    # Is this tag for a function/method?
                    if (eParseFunctions & s.modes) and (tag[TAG_TYPE] in ["function","method"]):
                        # See ParseFile() why line numbers are scaled.
                        digestedTokenDict[lineNumScaled] = (eTokenTag,tag[TAG_NAME]) # a token
                        # An empty line between helps the user to click.
                        digestedTokenDict[lineNumScaled+1] = (eTokenText,"()\n\n")
                    # Is this a tag for data?
                    if (eParseData & s.modes) and not (tag[TAG_TYPE] in ["function","method"]):
                        digestedTokenDict[lineNumScaled]   = (eTokenText,tag[TAG_TYPE] + " ")
                        digestedTokenDict[lineNumScaled+1] = (eTokenTag,tag[TAG_NAME])
                        digestedTokenDict[lineNumScaled+2] = (eTokenText,"\n\n")
        except:
            PrintException()
