# $Id: Tags.py,v 1.36 2006-03-05 05:02:34 devel Exp $
# pypersrc.
# Author:   Jim Brooks  http://www.jimbrooks.org
# LEGAL:    COPYRIGHT (C) JIM E. BROOKS 2005
#           THIS SOURCE CODE IS RELEASED UNDER THE TERMS
#           OF THE GNU GENERAL PUBLIC LICENSE (GPL).
# ==============================================================================

#-------------------------------------------------------------------------------
# Class Tags.
#-------------------------------------------------------------------------------

# Import module exported from C++ component.
import pycpp_tags

#-------------------------------------------------------------------------------
# A source code tag (not to be confused with a Tk text tag) is a tuple
# consisting of (name,type,file,line) that originates from Exuberant ctags.
# Phony tags are produced by pypersrc itself which are shown as "results"
# in a text widget.  Eg the results from a text search, shown as hyperlinks,
# are phony tags.
#-------------------------------------------------------------------------------

# Tag tuple indexes.
TAG_NAME = 0
TAG_TYPE = 1
TAG_FILE = 2
TAG_LINE = 3

class Tags(object):
    """ This Python class uses the integrated Exuberant ctags and
        C++ components to tag source code. """

    def __init__( s, argList ):
        s.ProcessArgsPre( argList )
        s.mutex = threading.RLock()   # reentrant lock
        
        # ctags uses the env var TMPDIR.
        tmpdir = "/tmp"
        if os.environ.has_key( "TMPDIR" ):
            tmpdir = os.environ["TMPDIR"]
        else:
            # $HOME/tmp exists?
            if os.environ.has_key( "HOME" ):
                dir = os.environ["HOME"] + "/tmp"
                if os.path.isdir( dir ):
                    tmpdir = dir
        os.environ["TMPDIR"] = tmpdir
        
        s.ProcessArgsPost( argList )

    def ProcessArgsPre( s, argList ):
        """ Assigns None to elements in argList that were processed. """
        return True # NOP

    def ProcessArgsPost( s, argList ):
        """ Assigns None to elements in argList that were processed. """
        return True # NOP

    def TagStatusbarUpdate( s ):
        try:
            s.mutex.acquire()    # protect s.taggedFileCnt
            st.gui.statusbar.Print( "Tagging: " + str(s.TagCnt()) + " tags in " \
                                    + str(s.taggedFileCnt) + "/" + str(s.srcFileCnt) + " files", \
                                    baton=True )
        finally:
            s.mutex.release()
    
    def TagFiles( s, fileList ):
        """ Pass filenames to Exuberant ctags.
            fileList --> Exuberant ctags --> tag lines --> C++ multimap. """
        st.log.Write( "Tags: starting" )
        s.FreeTags( fileList )    # free tags before re-tagging due to refresh
        st.gui.statusbar.SetUpdateCallback( s.TagStatusbarUpdate )
        s.taggedFileCnt = 0
        s.srcFileCnt = len(fileList)
        for file in fileList:
            if st.escKey:
                st.log.Write( "Tags: stopped by user" )
                break
            pycpp_tags.TagFile( file )
            s.mutex.acquire()    # protect s.taggedFileCnt
            s.taggedFileCnt += 1
            s.mutex.release()
        st.log.Write( "Tags: total = " + str(s.TagCnt()) )
        st.gui.statusbar.SetUpdateCallback( None )
        return

    def TagCnt( s ):
        """ Return amount of tags. """
        return pycpp_tags.TagCnt()

    def FreeTags( s, fileList ):
        """ Free tags belonging to source code files. """
        for file in fileList:
            pycpp_tags.FreeTags( file )
        return

    def IfTagRef( s, tagName, fileName, line ):
        """ Return true if this is a reference to a tag. """
        # For example, a reference to a tag should be drawn as a hyperlink.
        # But a tag should be drawn normally on the line where it is defined.
        return pycpp_tags.IfTagRef( tagName, fileName, line )

    def FindTag( s, tagName, tagType=None, tagFile=None, caseSens=True, partial=False, regex=False ):
        """ Return a list of matching tags or None. """
        # Search for a tag can be constrained by tag's type and/or tag's file.

        # Match tags by name.
        tagList = pycpp_tags.FindTag( tagName, caseSens, partial, regex )
        
        # Constrain by tag type (partial matching).
        if tagType:
            tagList2 = []
            for tag in tagList:
                thisTagType = tag[TAG_TYPE]
                if thisTagType.find( tagType ) != -1:
                    tagList2.append( tag )
            tagList = tagList2
        
        # Constrain by tag file (fileglob matching).
        if tagFile:
            tagList2 = []
            for tag in tagList:
                thisTagFile = tag[TAG_FILE]
                if MatchFileglob( thisTagFile, tagFile ):
                    tagList2.append( tag )
            tagList = tagList2
        
        return tagList

    def GetModuleTags( s, moduleName ):
        """ Return a list of tags in a module (source file). """
        return pycpp_tags.GetModuleTags( moduleName )

    def ChooseTagCmp( s, choice1, choice2 ):
        # sort() comparison callback.  Used by ChooseTag().
        score1 = choice1[0]
        score2 = choice2[0]
        if score1 < score2:
            return 1          # 1 (not -1) so that highest score will be first element
        elif score1 == score2:
            return 0
        else:
            return -1

    def ChooseTag( s, clickedTagName, fnameBrowse ):
        """ Return a good and bad list of matching tags. """
        # This is called when the user clicks a hyperlink to a src code tag.
        # In the good list, the best match is the first element
        # and the least likely is last.
        # The bad list shows tags which are probably bad matches
        # such as a matching tag in a different programming language file. """
        # For tags that have the same name, the best matching tag would be:
        # - defined in the same file the user is browsing
        # - or the same kind of programming language
        # - or its pathname is nearest the file the user is browsing
        # The tags are ranked by accumulating a score:
        # 100000 : defined in the same file the user is browsing
        #  10000 : the same kind of programming language
        # -10000 : different kind of programming language
        #  <1000 : pathname is nearest the file the user is browsing
        #          (count of matching chars in pathname)
        # Then the good list is sorted by scores.
        # TODO:
        # - Rank higher methods in the same class.
        # - Rank higher a tag closer to where the user clicked.
        
        tagList = s.FindTag( clickedTagName )
        if not tagList:
            return ([],[])
        elif len(tagList) == 1:
            return (tagList,[]) # only one tag matches
        goodList = []
        badList  = []
        for tag in tagList:
            score = 0
            fname = tag[2]
            if fname == fnameBrowse:
                score += 100000
            if WhichKindOfSrcFile( fname ) == WhichKindOfSrcFile( fnameBrowse ):
                score += 10000
            else:
                score -= 10000
            score += s.CountMatchingChars( fname, fnameBrowse )
            t = (score,tag)
            if score > 0:
                goodList.append( t )
            else:
                badList.append( t )
                
        # Sort the good list so that the tag with highest score
        # is the first element (to be shown at top of results text widget).
        goodList.sort( s.ChooseTagCmp )
        
        # Derive tag lists from good/bad lists (sans scores).
        goodTagList = []
        for choice in goodList:
            tag = choice[1]
            goodTagList.append( tag )
        badTagList = []
        for choice in badList:
            tag = choice[1]
            badTagList.append( tag )
        return (goodTagList,badTagList)

    def CountMatchingChars( s, x, y ):
        """ From the left, count chars in two strings that match. """
        n = 0
        i = 0
        while i < len(x) and i < len(y):
            if x[i] == y[i]:
                n += 1
            i += 1
        return n
                
    def PhonyTag( s, tagName, fileName, lineNum ):
        """ Return a new phony tag. """
        return (tagName,"phony",fileName,lineNum)

    def IfPhonyTagInToken( s, token ):
        try:
            # Could be a better way.
            return token[0] == eTokenTag and token[1][TAG_TYPE] == "phony"
        except:
            return False
