# $Id: Threads.py,v 1.26 2006-03-26 03:03:32 devel Exp $
# pypersrc
# Author:   Jim Brooks  http://www.jimbrooks.org
# LEGAL:    COPYRIGHT (C) JIM E. BROOKS 2005
#           THIS SOURCE CODE IS RELEASED UNDER THE TERMS
#           OF THE GNU GENERAL PUBLIC LICENSE (GPL).
# Notes:
# ==============================================================================

# ==============================================================================
# Thread design:
#
# Why are multiple threads used?  To fix the annoyance of the statusbar/GUI
# not being updated during long operations by running them in another thread.
#
# pypersrc can run 1 or 2 threads using the Python "threading" module.
# Most GUI functions aren't thread-safe, so the technique of top/bottom halfs is used.
# The initial main thread is dedicated to the GUI (it can call any GUI method).
# Long operations (taking more than a few seconds) are run by a temporary second thread
# which is forbidden to call GUI methods except a few that are thread-safe.
#
# The second thread is created as an object of the Python class threading.
# It executes as a top-half and stores its results as members.
# The first thread polls the second from a GUI idle function.
# When the second thread expires, its threading object persists as a zombie.
# The first thread "harvests" the results by calling the threading object's BottomHalf()
# (even though the second thread of execution it was meant for has expired!).
# 
# Skeleton class for thread objects based on the idea of top/bottom halfs.
# The run method (required by the threading class) is what invokes the top-half.
# Threads.Spawn() creates a Thread* object and passes a set of args to it
# as if it were a function.  How the args are passed to the top-half
# is somewhat convoluted: the Thread* constructor stores them, and the run() method
# passes them to the TopHalf() method.
#
# class ThreadDerived(ThreadBase):
#    
#    def __init__( s, *argsTopHalf ):
#       ThreadBase.__init__( s, argsTopHalf )
#        
#    def TopHalf( s [,arg1,arg2,..] ):
#        # This is executed by the spawned thread.
#        ...
#        # returning terminates the thread
#
#    def BottomHalf( s ):
#        # This is called and run by the main thread (not this thread) at an idle function.
#
# Remember: Bottom-half code usually shouldn't contain any calls to the GUI.
# Calling st.gui.Method() is a bug unless it is thread-safe (such as Statusbar.Print()).
#
# ==============================================================================

#-------------------------------------------------------------------------------
# Class Mutex.
#-------------------------------------------------------------------------------

class MutexBase(object):
    """ Helper class providing a wrapper for executing a function protected by a mutex. """
    # Example:
    #     class Class(MutexBase):
    #         def CritSect( s, x, y):
    #         def Whatever( s ):
    #             s.MutexCall( s.CritSect, x, y )
    #                          ^ pass as self.method (not just method)

    def __init__( s ):
        s.mutex = threading.RLock()    # reentrant lock

    def MutexCall( s, method, *args ):
        """ Protect with a mutex a call to a class method having a critical section. """
        # This may look as if it could return before releasing the mutex,
        # but the finally clause will always execute in any case.
        try:
            s.mutex.acquire()
            return method( *args )    # in Python, functions always return something (None by default)
        finally:
            s.mutex.release()    # always executes (even if try returns)

#-------------------------------------------------------------------------------
# Class Threads (unrelated to ThreadBase).
#-------------------------------------------------------------------------------

class Threads(object):
    # Manages the second thread.
    # Threads are spawned by pass their class object to Spawn().
    # __Harvest() is called by the main thread when GUI is idle,
    # and executes the bottom-half of an expired thread object
    # (does something with the results the thread computed).

    def __init__( s, argList ): # argList: cmd-line args (unused)
        # PRECOND: GUI was started.

        # Points to the second thread which is an object of class Thread*.
        s.thread = None

    # INTERFACE
    def Spawn( s, ThreadClass, *argsTopHalf ):
        # Spawn a second thread.
        # Passes args to thread'ss top-half when thread is activated.
        # ThreadClass must be a pypersrc Thread* class (derived from threading class).

        if (s.thread is not None) and (s.thread.isAlive()):
            # For simplicity, limit pypersrc to a main thread and one temporary second thread.
            # s.thread may point to an expired object.
            PrintWarning( "thread still busy, try again" )
        else:
            # Call class constructor, instantiate a "threading" object.
            s.thread = ThreadClass( *argsTopHalf )  # construct threading object
            s.thread.start()                        # schedule second thread
            s.__ScheduleHarvest()                   # schedule GUI idle func (for main thread)

    # PRIVATE
    def __Harvest( s ):
        # This is called by the main thread when the GUI idle.
        # A second thread exists.  Is it still running?
        if s.thread is not None:
            if s.thread.isAlive():
                s.__ScheduleHarvest()    # still running, reschedule
            else:
                # Clear s.thread before (not after) calling BottomHalf()
                # as it could spawn a thread which sets s.thread.
                thread_ = s.thread
                s.thread = None
                thread_.BottomHalf()    # call bottom half

    # PRIVATE
    def __ScheduleHarvest( s ):
        st.gui.ScheduleTimerFunc( 250, s.__Harvest )

#-------------------------------------------------------------------------------
# Class ThreadsBase.
#-------------------------------------------------------------------------------

class ThreadBase(threading.Thread):
    
    def __init__( s, *argsTopHalf ):
        # This is the thread constructor.  The thread really isn't running yet.
        # Store the passed args which are actually for TopHalf(), not this.
        s.argsTopHalf = argsTopHalf
        threading.Thread.__init__( s )

    def run( s ):
        """ This is where the thread begins by threading.start(). """
        # Pass the stored args to TopHalf().
        s.TopHalf( *s.argsTopHalf )

    def TopHalf( s ):
        """ Thread executes this.  Returning causes it to terminate. """
        # Derived class defines TopHalf()'s args.
        raise Oops( "abstract method" )

    def BottomHalf( s ):
        """ This is called and run by the main thread (not this thread) at an idle function. """
        raise Oops( "abstract method" )

#-------------------------------------------------------------------------------
# Class ThreadGatherTag
#-------------------------------------------------------------------------------

class ThreadGatherTag(ThreadBase):
    
    def __init__( s, *argsTopHalf ):
        ThreadBase.__init__( s, *argsTopHalf )
        s.srcFileList = []
        
    def TopHalf( s, argList ):  # argList: cmd-line args
        # Gather and tag source files.
        s.srcFileList = st.gather.Gather( argList )
        if s.srcFileList:
            st.gui.statusbar.Print( "Tagging source files...", timeout=0 )
            st.tags.TagFiles( s.srcFileList )
            # ...the rest is deferred to BottomHalf()...
        else:
            st.gui.statusbar.Print( "You need to pass names of source files or a directory (nothing to browse)." )

    def BottomHalf( s ):
        """ This is called and run by the main thread (not this thread) at an idle function. """
        st.gui.SetFileList( s.srcFileList )
        st.gui.statusbar.Print( st.GetSplashText() )
        if not s.srcFileList:
            text = "No src files were gathered.\nPass names of dirs/files on cmd-line."
            PrintWarning( text )

#-------------------------------------------------------------------------------
# Class ThreadRefresh
#-------------------------------------------------------------------------------

class ThreadRefresh(ThreadBase):
    
    def __init__( s, *argsTopHalf ):
        ThreadBase.__init__( s, *argsTopHalf )
        
    def TopHalf( s ):
        # Regather src files to determine if refresh is necessary.
        st.gui.statusbar.Print( "Refreshing...", timeout=0, now=True )
        st.log.Write( "refresh: start" )
        (s.change, s.modifiedList, s.removedList, s.addedList, s.unchangedList) = st.gather.Refresh()
        if s.change:
            st.refreshCnt += 1    # only increment if a change
            st.log.Write( "refresh: source files changed" )
            if s.modifiedList:
                st.log.WriteList( s.modifiedList, prologue="refresh: modified file: " )
            if s.removedList:
                st.log.WriteList( s.removedList, prologue="refresh: removed file: " )
            if s.addedList:
                st.log.WriteList( s.addedList, prologue="refresh: added file: " )

    def BottomHalf( s ):
        """ This is called and run by the main thread (not this thread) at an idle function. """        
        # Did src files change?
        if s.change:
            # Refresh.
            st.gui.Enable( False )
            try:
                s.Refresh()
            except:
                PrintException()
            st.gui.Enable( True )
        else:
            # No refresh necessary.
            st.log.Write( "refresh: no src files changed" )
        st.log.Write( "refresh: done" )

    def Refresh( s ):
        """ Refresh after src files has changed. """
        # PRECOND: Src files changed
        # PRECOND: GUI is disable.

        # Refresh tags.
        st.gui.statusbar.Print( "Refreshing tags...", timeout=0 )
        freeTagFileList = s.modifiedList + s.removedList
        tagFileList     = s.modifiedList + s.addedList
        st.tags.FreeTags( freeTagFileList )
        if tagFileList:
            st.log.Write( "refresh: tagging modified/added files" )
            st.tags.TagFiles( tagFileList  )
        else:
            st.log.Write( "refresh: tags = " + str(st.tags.TagCnt()) )
        # Refresh history.
        st.gui.resultsHistory.Refresh( s.change, s.addedList, s.modifiedList, s.removedList )
        # Refresh source file listbox.
        srcFileList = s.modifiedList + s.unchangedList + s.addedList
        srcFileList.sort()
        st.gui.SetFileList( srcFileList )
        # Flush affected src files out of cache.
        for fname in s.modifiedList + s.removedList:
            st.gui.cacheSrcFile.Flush( fname )
        # Refresh text widgets showing affected src files.
        for textWid in st.gui.notebookSrcFile.GetTextWidList():
            textWid.Refresh( s.change, s.addedList, s.modifiedList, s.removedList )
        # Refresh the results text widget which may be showing stale results.
        # Subtle: ResultsTextWid.Refresh() can cause a second thread to spawn
        # (eg global search) which will work as this is run by the main thread.
        st.gui.GetResultsTextWid().Refresh( s.change, s.addedList, s.modifiedList, s.removedList )
        st.gui.statusbar.Print( "Refreshing...", timeout=1000 )   # timeout in one second

#-------------------------------------------------------------------------------
# Class ThreadFindText
#-------------------------------------------------------------------------------

class ThreadFindText(ThreadBase):
    
    def __init__( s, *argsTopHalf ):
        ThreadBase.__init__( s, *argsTopHalf )
        
    def TopHalf( s, command, finder, srcFileList ):
        # Top-half:
        # This produces HTML-like results in a text widget.
        # The user can click on a hyperlink to jump to the match.
        # This builds a token list consisting of two types of tuples:
        # (eTokenTag,(PhonyTag))
        # (eTokenText,text)

        tokenList = []
        pattern = finder.GetPattern()
        local = (len(srcFileList) == 1)

        # Although searchedFileCnt is accessed by two threads, no mutex
        # is necessary as it's just a fundamental type used only for display.
        searchedFileCnt = 0
        totalFileCnt = len(srcFileList)
        def SearchUpdate():
            st.gui.statusbar.Print( "Searching for `" + pattern + "` in " \
                                    + str(searchedFileCnt) + "/" + str(totalFileCnt) + " files", \
                                    baton=True )
        st.gui.statusbar.SetUpdateCallback( SearchUpdate )
        
        for srcFile in srcFileList:
            if st.escKey:
                st.log.Write( "ThreadFindText: searching stopped by user" )
                break
            
            searchedFileCnt += 1

            # Find all matches in this src file.
            try:
                text = ReadSrcFile( srcFile )
            except IOError:
                text = ""
                PrintWarning( "ThreadFindText: can't open file " + srcFile )
            lineList = finder.GetMatchingLines( text, srcFile )
            for line in lineList:
                phonyTag = st.tags.PhonyTag( srcFile + ":" + str(line), srcFile, line )
                tokenList.append( (eTokenTag,phonyTag) )

        st.gui.statusbar.SetUpdateCallback( None )

        if len(srcFileList) == 1:
            # fnames of local searches should be unique.
            fname = "[results: " + command + " (" + srcFileList[0] + ") ]"
        else:
            fname = "[results: " + command + "]"
        finderModes = finder.GetModesString()
        if tokenList:
            tokenList.insert( 0, (eTokenText,"Found `" + pattern + "` " + finderModes + ":\n\n") )
        else:
            tokenList.insert( 0, (eTokenText,"Did not find `" + pattern + "` " + finderModes + ".\n\n") )

        # Save state for bottom-half.
        s.tokenList = tokenList
        s.fname = fname
        s.command = command
            
    def BottomHalf( s ):
        """ This is called and run by the main thread (not this thread) at an idle function. """
        contents = ContentsSearch( tokenList=s.tokenList, fname=s.fname, command=s.command )
        st.gui.GetResultsTextWid().LoadContents( contents=contents )
        st.gui.statusbar.Clear()
