# $Id: lib.py,v 1.23 2006-03-03 13:42:02 devel Exp $
# Utility functions.
# Author:   Jim Brooks  http://www.jimbrooks.org
# LEGAL:    COPYRIGHT (C) JIM E. BROOKS 2005
#           THIS SOURCE CODE IS RELEASED UNDER THE TERMS
#           OF THE GNU GENERAL PUBLIC LICENSE (GPL).
# ==============================================================================

import os, sys, stat, string, fnmatch

#-------------------------------------------------------------------------------
# Exceptions.
#-------------------------------------------------------------------------------

import traceback

def PrintCallStack():
    """ Print call stack. """
    traceback.print_exc( file=sys.stderr )

def PrintException( exception=None ):
    """ Print information when an exception occurs. """
    """ Historically, PrintException() with no arg was placed in an "except:" clause. """
    """ However, the recommended new usage is to pass an exception object. """

    if ( exception == None ):
        print( "EXCEPTION: " )
    else:
        try:
            print( "EXCEPTION: \"" + str(exception) + "\"" )
        except:
            print( "EXCEPTION: " )

    PrintCallStack()

class Oops(Exception):  # class Oops(object) would be incorrect
    """ User-defined exception class. """
    """ """
    """ Example: """
    """ try: """
    """    raise Oops( "file missing" ) """
    """ except Exception as exception:  # class as instance """
    """    PrintException( exception ) """
    """ """
    """ Note: 'except class as instance' is the new Python 2.6/3.x syntax. """
    
    def __init__( s, text="" ):
        s.text = text[:]
        
    def __repr__( s ):
        return s.text

    def __str__( s ):  # need both __repr__ and __str__ despite Python docs
        return s.text

#-------------------------------------------------------------------------------
# Basic functions.
#-------------------------------------------------------------------------------

def Unique( seq ):
    """ Remove duplicates out of a sequence.  Returns a new sequence (UNDEFINED ORDER). """
    # This terse trick converts to a set then back to the sequence type.
    # A set intrinsincally eliminates duplicates.
    return seq.__class__( set( seq ) )

def List2String( list, separator="" ):
    s = ""
    for x in list:
        s = s + x + separator
    return s

def DeleteFromList( list, functor ):
    """ Remove elements from a list that match a criteria. """
    """ Callback will be passed every element which should return True """
    """ if the element should be deleted. """
    
    mfd = []   # marked for deletion
    
    i = 0
    for elem in list:
        if functor( elem ):
            mfd.append( i )
        i += 1

    # This is a little tricky: when one element is deleted,
    # all the remaining indexes become 1 too great and so on.
    i = 0
    deletions = 0
    for idx in mfd:
        del list[idx-deletions]
        deletions += 1

def DeleteFromDict( dict, functor ):
    """ Remove key/value pairs from a dictionary that match a criteria. """
    """ functor is passed every key which should return True """
    """ if the key/value pair should be deleted. """
    
    mfd = []  # marked for deletion
    
    for key in dict.keys():
        if functor( key ):
            mfd.append( key )

    for key in mfd:
        del dict[key]

#-------------------------------------------------------------------------------
# String functions.
#-------------------------------------------------------------------------------

def IfEol( c ):
    """ Return True if end-of-line. """
    return c == "\n"

def Chomp( s ):
    """ Discard last char if it's newline """
    if s[-1] == "\n":
        s = s[0:-1]
    return s

def IfQuoted( text ):
    """ Return true if string is quoted. """
    return len(text) >= 2 and text[0] in "\"'" and text[-1] in "\"'"

def Unquote( text ):
    """ Removes the 2 surrounding quote chars out of a string. """
    # PRECOND: String is quoted (no checks).
    return text[1:-1]    # chop first and last char

def UnquoteWords( wordList ):
    """ Remove surrounding quote chars in a list of strings. """
    i = 0
    for word in wordList:
        if IfQuoted( word ):
            wordList[i] = Unquote( word )
        i += 1
    return wordList

def SplitWords( wordStr, keepOuterQuotes=True ):
    """ Split a string into a list but does not split quoted words. """
    wordList = []   # result
    word = ""       # accumulates a word char-by-char
    quote = None    # holds { "'", '"', None }
    whitespace = " \t\n"
    i = 0
    while i < len(wordStr):
        c = wordStr[i]
        # Escaped char?
        if c == "\\":
            word += c
            if i+1 < len(wordStr):
                word += wordStr[i+1]
                i += 1
        # Quote char?
        elif c == '"' or c == "'":
            # Beginning quote char?
            if not quote:
                if keepOuterQuotes:
                    word += c
                # Remember which kind of quote char to determine final quote char.
                quote = c
                # A pair of the same quote char?
                if i+1 < len(wordStr) and wordStr[i+1] == wordStr[i]:
                    word += c
                    i += 1
            else:
                # A pair of the same quote char?
                if i+1 < len(wordStr) and wordStr[i+1] == wordStr[i]:
                    if keepOuterQuotes:
                        word += c
                    word += c
                    if c == quote:   # to handle '""Hello World""'
                        quote = None
                    i += 1
                # Ending quote char?
                elif c == quote:
                    if keepOuterQuotes:
                        word += c
                    quote = None
                # Different kind of quote char than beginning one.
                else:
                    word += c
        # Quoting?
        elif quote:
            # Append this char before the space case below.
            word += c
        # Whitespace?
        elif c in whitespace:
            # Append accumulated word (unless empty) to list.
            if word:
                wordList.append( word )
                word = ""
        else:
            word += c
        # Continue to next char.
        i += 1

    # Append final word (unless empty) to list.
    if word:
        wordList.append( word )
        
    return wordList

def IfHaveEachChar( set1, set2 ):
    """ Return true if each char in set1 exists in set2. """
    for c1 in set1:
        if not c1 in set2:
            return False
    return True

def AddIntoParentheses( parentheses, addition ):
    """ Helper to add comma-separated sub-strings inside a parenthesized string. """
    # Eg: "(abc)" can become "(abc,def)"
    if parentheses == "":
        parentheses = "(" + addition + ")"
    else:
        parentheses = parentheses[:-1] + "," + addition + ")"
    return parentheses

#-------------------------------------------------------------------------------
# File I/O.
#-------------------------------------------------------------------------------

def ReadLine( fd ):
    """ Generator function to read lines from a file descriptor.
        Returns the line excluding the newline char. """
    remainder = ""
    while True:
        piece = os.read( fd, 16384 )
        # Continue with remainder (that didn't form a complete line).
        piece = remainder + piece
        if not piece:
            return
        remainder = "" # consumed
        # Is the last line a complete line?
        if piece[-1] != "\n":
            # Is the incomplete line the only line?
            # If so, wait for more input.
            if piece.find( "\n" ) == -1:
                remainder = piece
                continue
            else:
                # Extract out the incomplete last line into "remainder".
                i = piece.rfind( "\n" )
                try:
                    remainder = piece[i+1:]
                except:
                    remainder = ""
                piece = piece[:i]
        # Now yield all of the complete lines.
        lineList = piece.split( "\n" )
        if not lineList[-1]:
            del lineList[-1]    # split() will make the last element an empty line
        for line in lineList:
            yield line[:-1]    # minus newline char
    return

def WriteBuffer( fd, buf ):
    """ Write every byte of a buffer to a file descriptor.
        This is necessary for pipes, sockets. """
    buf2 = buf[:]
    while len(buf2):
        n = os.write( fd, buf2 )
        buf2 = buf2[n:]

def IsExecutableFile( fname ):
    """ True if is a file and is executable (returns false for directories). """
    status = os.stat( fname )
    mode = status[0]
    if (mode & stat.S_IEXEC) and not (mode & stat.S_IDIR):
        return True
    return False

def FileModificationTime( pname ):
    """ Get a file's modification time.  Exception if file doesn't exist. """
    t = os.stat( pname )
    return t[stat.ST_MTIME]

#-------------------------------------------------------------------------------
# Pathnames.
#-------------------------------------------------------------------------------

def FileExt( fname ):
    """ Return a file's extension/suffix (including '.') or None. """
    try:
        i = fname.rindex( "." )
        return fname[i:]
    except:
        return None

def PrettyDirName( dirname ):
    """ Return dirname with a trailing slash. """
    try:
        if dirname[-1] != "/":
            dirname += "/"
        else:
            # In case another trailing slash was blindly appended.
            if dirname[-2:] == "//":
                dirname = dirname[:-1]
    except:
        PrintException()
    return dirname

def AbsPath( fname, cwd=None ):
    """ Return an absolute path. """
    if not cwd:
        cwd = os.getcwd()
    if IfCygwin():
        # Change to cygwin naming convention: /cygdrive/c/cwd/fname
        if len(fname) > 2 and fname[1] == ":":   # DOS drive letter?
            return AbsPathDos2Cygwin( fname )
        else:
            fname = fname.replace( "\\", "/" )
            return os.path.join( cwd, fname )
    else:
        if fname[0] == "/":
            return fname
        else:
            return os.path.join( cwd, fname )

def AbsPathDos2Cygwin( path ):
    path = "/cygdrive/" + path[0] + "/" + path[2:]
    path = path.replace( "\\", "/" )
    return path.replace( "//", "/" )

def TransformPath( path ):
    """ Cope with differences in OS file names. """
    if not IfCygwin():
        return path
    else:
        # cygwin
        savedPath = path
        try:
            if path.find( "/cygdrive" ) == 0:
                # /cygdrive/c --> c:\
                driveLetter = path[10]
                path = driveLetter + ":" + path[11:]
                path = path.replace( "/", "\\" )    # to MSDOG dir separator
            return path
        except:
            return savedPath

def TransformPathList( pathList ):
    """ Cope with differences in OS file names. """
    i = 0
    while i < len(pathList):
        pathList[i] = TransformPath( pathList[i] )
        i += 1

def Basename( fname ):
    # os.path.basename doesn't work right on cygwin.
    try:
        i1 = fname.rfind( "/" )
        i2 = fname.rfind( "\\" )
        if i1 == -1 and i2 == -1:
            return fname
        if i1 > i2:
            return fname[i1+1:]
        else:
            return fname[i2+1:]
    except:
        return fname

def MatchFileglob( fname, glob, caseSens=False ):
    """ Return true if fname matches glob pattern. """
    # This wrapper exists because the behavior of fnmatch.fnmatch()
    # is inconsistent across platforms (it may or may not be case-sensitive).
    # fnmatchcase() is supposed to be always case-sensitive.
    if not caseSens:
        fname = fname.lower()
        glob  = glob.lower()
    return fnmatch.fnmatchcase( fname, glob )    # fnmatchcase() is always case-sensitive

#-------------------------------------------------------------------------------
# Process execution.
#-------------------------------------------------------------------------------

def RunCmd( fname ):
    """ Run a command, return its stdout as a string. """
    pipe = os.popen( fname + " 2>&1" )
    out = ""
    while 1:
        line = pipe.readline()
        if not line:
            break
        out = out + line
    pipe.close()
    return out

def RunProgram( fname, child, argTuple ):
    """ Run a program. """
    # If child is false, the Python process is replaced (exec).
    if IfCygwin():
        os.spawnv( os.P_NOWAIT, fname, argTuple )
        return
    if child:
        pid = os.fork()
        if pid != 0:
            # This is the parent.
            return
    try:
        if fname[0] == '/':
            os.execv( fname, argTuple )  # full pathname
        else:
            os.execvp( fname, argTuple ) # filename via PATH
    except:
        PrintWarning( "RunProgram(" + fname + ") failed" )
    # exec() failed.
    if child:
        # Stop child process.
        os._exit(0)  # not exit(0)

#-------------------------------------------------------------------------------
# Operating-system specific.
#-------------------------------------------------------------------------------

# Be careful of ordering calls to these when determining which OS.
# IfUnix() is very generic and somewhat of a catch-all.

def IfCygwin():
    return sys.platform.find( "cygwin" ) != -1

# Mac OS X
def IfDarwin():
    return sys.platform.find( "darwin" ) != -1

def IfUnix():
    # python/lib/module-os.html:
    # The following names have currently been registered:
    # 'posix', 'nt', 'mac', 'os2', 'ce', 'java', 'riscos'
    return os.name == "posix"
        
#-------------------------------------------------------------------------------
# Finder class.
#-------------------------------------------------------------------------------

import re

class OopsBadRegex(Oops):
    def __init__( s, text="" ):
        Oops.__init__( s, text )

class Finder(object):
    """ Class to facilitate text searching using various matching methods. """

    def __init__( s, pattern, caseSens=False, exact=False, regex=False, glob=False, negative=False ):

        # Construct according to search method.
        # - method: regex
        # - method: file glob matching
        # - method: basic string matching
        # - method: exact
        # - case-sensitive (all methods)
        # - negative: IfMatch() returns True not pattern does not match
        s.patternOrig = pattern[:]      # copy
        s.pattern     = pattern
        s.caseSens    = caseSens        # true if case-sensitive
        s.regex       = regex
        s.exact       = exact
        s.glob        = glob
        s.negative    = negative
        s.basic       = not (exact or regex or glob)

        # The user is supposed to surround a multi-word pattern with quotes.
        # The surrounding quotes should be discarded.  Which leads to the question:
        # How can the user include a quote char in the pattern?
        # What if the pattern is '"Hello World"'?  The question answers itself:
        # by surrounding the pattern with two pairs of quotes at both ends.
        if IfQuoted( s.pattern ):
            s.pattern = Unquote( s.pattern )    # removes the 2 outermost quotes only

        # Compile regex.
        try:
            if s.regex:
                if s.caseSens:
                    s.regexObj = re.compile( s.pattern )
                else:
                    s.regexObj = re.compile( s.pattern, re.IGNORECASE )
        except:
            # This can happen if the user enters an unquoted multi-word regex
            # so that only the first part of the regex reaches here.
            raise OopsBadRegex

    def IfMatch( s, haystack ):
        if not s.negative:
            return s.__IfMatch( haystack )
        else:
            return not s.__IfMatch( haystack )
            
    def __IfMatch( s, haystack ):
        """ Return true if passed haystack matches pattern (given to constructor). """
        # The match methods have varying precedence or are mutually-exclusive.
        if s.basic:
            return Find( haystack, s.pattern, s.caseSens ) != -1
        elif s.glob:
            return MatchFileglob( haystack, s.pattern, caseSens=s.caseSens )
        elif s.regex:
            return s.regexObj.search( haystack )
        elif s.exact:
            return s.pattern == haystack
        else:
            raise Oops( "Finder.IfMatch: unhandled search method" )

    def GetMatchingLines( s, haystack, fnameHaystack="" ):
        """ Return a list of line numbers that match pattern. """
        # fnameHaystack is informational (shown in error message, no file is read here).
        if s.negative:
            raise Oops( "unhandled case" )
        if s.basic:
            # This is a common case written in C++ for speed.
            return FindLines( haystack, s.pattern, s.caseSens )
        else:
            # Search thru file before searching line-by-line (usually saves time).
            # Rare case: if a data file has a src file ext, Python will raise TypeError
            # while parsing the args for Find() (written in C++ at Utils.cc).
            matchingLines = []
            try:
                somewhere = s.IfMatch( haystack )
            except:
                somewhere = False
                if haystack.find( "\0" ):
                    PrintWarning( "Finder: found null char in " + fnameHaystack )
                else:
                    raise
            if somewhere:
                # Search line-by-line.
                lineNum = 1
                lineList = haystack.split( "\n" )
                for line in lineList:    # line is a string but lineNum is an int
                    if s.IfMatch( line ):
                        matchingLines.append( lineNum )
                    lineNum += 1
            return matchingLines            

    def GetPattern( s ):
        """ Returns specified pattern. """
        return s.patternOrig    # not s.pattern which could be modified internally

    def IfGlob( s ):
        return s.glob

    def IfRegex( s ):
        return s.regex

    def IfCaseSens( s ):
        return s.caseSens

    def GetModesString( s ):
        """ Returns finder's modes as an informational/printable string. """
        modes = ""
        if s.caseSens:
            modes = AddIntoParentheses( modes, "case-sensitive" )
        if s.regex:
            modes = AddIntoParentheses( modes, "regex" )
        if s.glob:
            modes = AddIntoParentheses( modes, "glob" )
        if s.exact:
            modes = AddIntoParentheses( modes, "exact" )
        if s.basic:
            modes = AddIntoParentheses( modes, "basic" )
        return modes

#-------------------------------------------------------------------------------
# Nest class.
#-------------------------------------------------------------------------------

class Nest(object):
    # Helper class to nest complementary actions.
    # IfProceed() returns True when the depth indicates an operation should be executed.
    # action1 has no meaning here, its complement is action2 (actions have meaning to caller).
    # action1=True increments depth, action1=False decrements depth.
    # count(post):  action1:   action2:   proceed:  outer-level:
    # +1            True       False      True      True
    # +2            True       False      False     False (nested)
    # +1            False      True       False     False (nested)
    # 0             False      True       True      True

    def __init__( s, nestFirst=True ):
        # nestFirst tells IfProceed() to bypass changing depth
        # the first time it is called.  Eg, the first time a GUI is enabled
        # should be ignored, thereafter, enable/disable GUI operations will nest.
        s.depth     = 0
        s.first     = True
        s.nestFirst = nestFirst

    def IfProceed( s, action1 ):

        # First time called?
        if s.first:
            s.first = False
            # Should the first call affect depth?
            if not s.nestFirst:
                return True     # leave depth alone, tell caller to proceed
        
        if not action1:
            s.depth -= 1
            if s.depth != -1:
                return False    # at inner level
        else:
            s.depth += 1
            if s.depth != 0:
                return False    # at inner level
            
        return True             # at outer level
