# $Id: misc.py,v 1.27 2006-03-05 15:56:57 devel Exp $
# Misc functions for pypersrc.
# Author:   Jim Brooks  http://www.jimbrooks.org
# LEGAL:    COPYRIGHT (C) JIM E. BROOKS 2005
#           THIS SOURCE CODE IS RELEASED UNDER THE TERMS
#           OF THE GNU GENERAL PUBLIC LICENSE (GPL).
# ==============================================================================

#-------------------------------------------------------------------------------
# Functions for source files, programming languages.
#-------------------------------------------------------------------------------

# Programming languages recognized by the  -l arg.
progLangList = [ "c", "c++", "lua", "python", "java", "perl", "forth" ]

progLangExtDict = { ".c"    : "c",
                    ".h"    : "c",
                    ".cc"   : "c++",
                    ".C"    : "c++",
                    ".cpp"  : "c++",
                    ".cxx"  : "c++",
                    ".c++"  : "c++",
                    ".H"    : "c++",
                    ".hh"   : "c++",
                    ".hxx"  : "c++",
                    ".h++"  : "c++",
                    ".inl"  : "c++",
                    ".java" : "java",
                    ".jav"  : "java",
                    ".j"    : "java",
                    ".lua"  : "lua",
                    ".py"   : "python",
                    ".pyw"  : "python",
                    ".pl"   : "perl",
                    ".pm"   : "perl",
                    ".ph"   : "perl",
                    ".4th"  : "forth",
                    ".fth"  : "forth",
                    ".seq"  : "forth",
                    ".fs"   : "forth", # GForth
                    }

def WhichKindOfSrcFile( pname ):
    """ Return which programming language a source file is for.
        Eg, returns 'c' for main.c.  Returns empty string if unknown. """
    (root,ext) = os.path.splitext( pname )
    if ext and progLangExtDict.has_key( ext ):
        return progLangExtDict[ext]
    else:
        return ""

def ReadSrcFile( fname ):
    """ Return string containing a source file or raise IOError exception. """
    f = open( fname, "r" )    # may raise exception
    buf = f.read()
    buf = buf.replace( "\r", "" )    # remove MSDOS CR chars
    f.close()
    return buf

def SrcFile2TokenList( fname ):
    """ Compute a token list from a source file. """
    # Read the source code file.
    # Lex file to build a token list.
    # Parser will build a second token list from the first.
    # Then load the text widget from the second.
    try:
        text = ReadSrcFile( fname )
    except IOError:
        PrintWarning( "SrcFile2TokenList: source file " + fname + " can't be read, continuing" )
        return ([],False)
    text = text.expandtabs( st.tabWidth )
    tokenList = st.lex.LexFile( fname, text )
    tokenList = st.parser.ParseFile( fname, tokenList )
    return (tokenList,True)

#-------------------------------------------------------------------------------
# Printing messages.
#-------------------------------------------------------------------------------

# "Warning" is defined by Python BTW.

def Print( *args ):
    """ Only print a message if verbose. """
    if st.verbose:
        __Print( sys.stdout, *args )

def PrintError( *args ):
    """ Print an error message. """
    st.error = True
    __Print( sys.stderr, "ERROR: ", *args )
    try:
        if st.gui.IfReady():
            st.gui.statusbar.PrintError( *args )
    except:
        pass

def PrintWarning( *args ):
    """ Print a warning message. """
    st.warning = True
    __Print( sys.stderr, "WARNING: ", *args )
    try:
        if st.gui.IfReady():
            st.gui.statusbar.PrintWarning( *args )
    except:
        pass

# PRIVATE
def __Print( file, *args ):
    """ Print to a file and the log. """
    try:
        logText = ""
        for a in args:
            astr = str(a)
            logText += astr
            file.write( astr )
            file.flush()
        if astr[-1] != "\n":
            logText += "\n"
            file.write( "\n" )
            file.flush()
        st.log.Write( logText )
    except:
        PrintException()

#-------------------------------------------------------------------------------
# Class Log.
#-------------------------------------------------------------------------------

class Log(object):
    # The usual methods are thread-safe.
    
    VERBOSE = -1
    NORMAL  = 0
    URGENT  = 1
    
    def __init__( s ):
        s.buf        = ""
        s.enabled    = True
        s.level      = Log.NORMAL
        s.splashLine = ""
        s.mutex      = threading.RLock()  # reentrant lock

    # INTERFACE
    def Clear( s ):
        s.mutex.acquire()
        s.buf = ""
        s.buf += s.splashLine
        s.mutex.release()

    # PRIVATE
    def __Write( s, text, level=NORMAL ):  # level=Log.NORMAL won't compile
        # PRECOND: Mutex was acquired.
        if s.enabled and level >= s.level:
            if text[-1] == "\n":
                s.buf += text
            else:
                s.buf += text + "\n"

    # INTERFACE
    def Write( s, text, level=NORMAL ):  # level=Log.NORMAL won't compile
        try:
            s.mutex.acquire()
            s.__Write( text, level )
        finally:
            s.mutex.release()
        
    # INTERFACE
    def WriteList( s, textList, level=NORMAL, prologue="" ):
        """ Write a list of log messages. """
        try:
            s.mutex.acquire()
            for text in textList:
                s.__Write( prologue + text, level )
        finally:
            s.mutex.release()

    # INTERFACE
    def Read( s ):
        """ Returns a copy of the log. """
        # Return a copy to prevent inteference by later logging.
        s.mutex.acquire()
        buf2 = s.buf[:]
        s.mutex.release()
        return buf2

    # INTERFACE
    def Enable( flag=True ):
        # Pass False to disable logging.
        # Returns the old flag so that it can be restored.
        s.mutex.acquire()
        old = s.enabled
        s.enabled = flag
        s.mutex.release()
        return old
        
    # INTERFACE
    def SetLogLevel( s, level ):
        s.level = level
        
    # INTERFACE
    def SetSplashLine( s, line ):
        """ Sets the first line shown in the log. """
        s.splashLine = line
        s.Write( s.splashLine )

#-------------------------------------------------------------------------------
# Easter eggs.
#-------------------------------------------------------------------------------

def Underdog():
    text = \
    "------------------------------------------------------------------------------\n"\
    " Look!  Up in the sky!  It's a bird!  It's a plane! \n"\
    " No, it's Underdog! \n"\
    "                                    ,.+----.\n"\
    "                               ,*''         I\n"\
    "              ____           /         Mbp. dP          __\n"\
    "          ,o@QBBBBBbm,_     ; ,d       /~`QMP'\\     _odBBBBbo,\n"\
    "          OBBROYALBBBBBBBBmOBBBM      |    |   |,*~dBBBBBBBBBBBb \n"\
    "          OBBBBBBBBBBqBBBBBBBBBP      |  0 |0 ,'  dBBBBBBBBBBBBBb\n"\
    "          '*BBBP'     ^OBBBP'          \\__/  ~    OBBBBBBBBBBBBBP\n"\
    "      /                     :           __.       `0BBBBBBBBBBP|\n"\
    "                    .-.     `          ~,A.         `0BBBBBBBBP |\n"\
    "                 .-(   `x__  `\\        .B*Bb          `'?OOP~   ;\n"\
    "              .-(   `\\/'   `.  `~--++*\\{  `Pb           ,'     ,'\n"\
    "   /         (   `\\_/'\\__/  )     _    !`-' !`.       ,'     ,'  \n"\
    "        /     `\\_/'       ,'     L '+,_|    |-,`.__ ,'    _,','~~\\.~~~\\\n"\
    "                 \\        /\\    / ',   '+.  |/ >   `~~~~~~  ,'    :/~  \\\n"\
    "    /             `~===\\/'  `\\/'    '+.   >{  /\\        __.,: |   ;    ;\n"\
    "           /           {         ,|    `\\/  \\/  `Y~~---~   `. `--:'~~~';\n"\
    "                        `\\     ,&#|    uuu    uuu \\         `,   ;^`~~';\n"\
    "        /               ,d#b,.&###|    UUU.   UUU. `\\     __.;~~' `~~~'\n"\
    "      /              ,d###########|    UUU;   |UU:   Y~~~~\n"\
    "                 ,d############i~      `UUUUUUUUU;   I  \n"\
    "           ,d##################I        `UUUUUUU'    I     /   /\n"\
    " (#############################b.                    Im \n"\
    "    )############################b.      \\           d#b\n"\
    "   (###############################b++g+++~#b      ,d##P\n"\
    "       }####################P~ ~Y####P    J#P`~~~T'\n"\
    "      (###################P'     `YP'    ;P'   ,'\n"\
    "        )###############P'        '    ,'    ,'    /\n"\
    "       (##############P'          __+~'   /##'\n"\
    "  /            Y####P'         ;-'     /###P\n"\
    "            /   Y#P'         /'      /&##P~\n"\
    "                 |         /'      /&#P~         \n"\
    "      /          `\\______/'\\_____/'            \n"\
    "                                    /            \n"\
    "                           /\n"\
    "              /                          \n"\
    "------------------------------------------------------------------------------"

    st.log.Write( text )
    print text
